import torch
import numpy as np
from arm_pytorch_utilities.make_data import select
import scipy.io as sio
import os


def gt_model(xu, params, M, s, selector=None):
    if selector is None:
        selector = select.identity_select
    m = selector(xu)
    y = xu.new_empty((xu.shape[0], M))
    for i in range(s):
        in_cluster = m == i
        y[in_cluster] = torch.mm(xu[in_cluster], params[i].view(-1, M))
    return y, m


def sort(U, Y, labels):
    I = torch.argsort(labels, dim=0)
    U = U[I]
    Y = Y[I]
    labels = labels[I]

    return U, Y, labels


def make_input(p, N, variance):
    U = np.random.rand(N, p) * 8 - 4
    U = np.concatenate((U, np.ones((N, 1))), 1)

    E = np.random.randn(N, 1) * np.sqrt(variance)

    return U, E


def make(params, N=200, variance=0.05, sorted=False, selector=None, filename=None):
    M = params.shape[2]
    s = params.shape[0]
    p = params.shape[1] - 1

    U, E = make_input(p, N, variance)
    E = torch.tensor(E, dtype=params.dtype, device=params.device)
    U = torch.tensor(U, dtype=params.dtype, device=params.device)

    Y, labels = gt_model(U, params, M, s, selector)
    Y += E

    if sorted:
        U, Y, labels = sort(U, Y, labels)

    if filename is not None:
        joined_data = {'XU': U.numpy(), 'Y': Y.numpy(), 'labels': labels.numpy()}
        try:
            sio.savemat(filename, joined_data)
        except FileNotFoundError:
            os.makedirs(os.path.dirname(filename))
            sio.savemat(filename, joined_data)

    return U, Y, labels


def load(filename, max_N=None, to_load=('XU', 'Y', 'labels')):
    prev_data = sio.loadmat(filename)

    loaded = [prev_data[l][:max_N] for l in to_load]
    tensors = [torch.tensor(ll) for ll in loaded]

    return tensors



import abc
import copy
from typing import Type

import torch
from arm_pytorch_utilities import load_data
from arm_pytorch_utilities import preprocess
from arm_pytorch_utilities import str_utils


class DataSource:
    """
    Source of data, agnostic to its creation method; public API for data
    """

    def __init__(self, max_N=None, config=load_data.DataConfig(), preprocessor: preprocess.Transformer = None,
                 device="cpu"):
        """
        :param N: number of data points for this data source; set to None for unknown/loaded
        :param config: configuration of data of this source
        :param preprocessor: data transformer/preprocessor, such as StandardizeVariance
        :param device:
        """
        self.max_N = max_N
        self.N = max_N
        self.config = config
        self._original_config = None

        self.d = torch.device(device)

        self._train = None
        self._val = None
        # data before preprocessing; set only if we have a preprocessor
        self._original_train = None
        self._original_val = None

        self.preprocessor = None  # implementation should use preprocessor in make_data
        self.update_preprocessor(preprocessor)

    def training_set(self, original=False):
        return self._original_train if original and self._original_train else self._train

    def validation_set(self, original=False):
        return self._original_val if original and self._original_val else self._val

    def original_validation_set(self):
        return self.validation_set(original=True)

    @abc.abstractmethod
    def make_data(self):
        """Create data that'll be available from training_set and validation_set"""

    def data_id(self):
        """String identification for this data"""
        return "N_{}".format(str_utils.f2s(self.N))

    def original_config(self):
        return self.config if self._original_config is None else self._original_config

    def current_config(self):
        return self.config

    def update_preprocessor(self, preprocessor):
        """Change the preprocessor,
        which involves remaking the data and potentially changing the meaning/config of the data.
        If you need the loaded untransformed config (for example for doing control in that space), then you should
        not pass the preprocessor in the constructor, but instead call this function with it afterwards.

        :param preprocessor:
        :return: original data configuration (before *any* transformation, even if this is not the first transform)
        """
        # the first time we're updating the preprocessor and the original data has been loaded
        if self._original_config is None and self.config.nx is not None:
            self._original_config = copy.deepcopy(self.config)

        self.preprocessor = preprocess.DatasetPreprocessor(preprocessor) if \
            isinstance(preprocessor, preprocess.Transformer) else preprocessor
        self.make_data()
        return self.original_config()


class FileDataSource(DataSource):
    def __init__(self, loader: load_data.DataLoader, data_dir, validation_ratio=0.2, **kwargs):
        """
        :param loader: data loader specializing to what's stored in each file
        :param data_dir: data directory or list of directories
        :param predict_difference: whether the output should be the state differences or states
        :param validation_ratio: amount of data set aside for validation
        :param kwargs:
        """
        self.loader = loader
        self._data_dir = data_dir
        self._validation_ratio = validation_ratio
        super().__init__(**kwargs)

    def make_data(self):
        full_set = load_data.LoaderXUYDataset(loader=self.loader, dirs=self._data_dir, max_num=self.max_N,
                                              config=self.config, device=self.d)
        train_set, validation_set = load_data.split_train_validation(full_set,
                                                                     validation_ratio=self._validation_ratio)

        self.N = len(train_set)

        if self.preprocessor:
            self.preprocessor.fit(train_set)
            self.preprocessor.update_data_config(self.config)
            # save old data (if it's for the first time we're using a preprocessor)
            if self._original_val is None:
                self._original_train = train_set[:]
                self._original_val = validation_set[:]
            # apply on training and validation set
            train_set = self.preprocessor.transform(train_set)
            validation_set = self.preprocessor.transform(validation_set)

        self._train = train_set[:]
        self._val = validation_set[:]

    def restrict_training_set_to_slice(self, restricted_slice):
        if self._train is not None:
            self._train = tuple(v[restricted_slice] for v in self._train)

    def data_id(self):
        """String identification for this data"""
        return "{}_N_{}_{}".format(self._data_dir, str_utils.f2s(self.N), self.config)



import re
from itertools import groupby

import torch


def discrete_array_to_value_ranges(discrete_array):
    """
    Return a sequence of (value, start, end) for an array of discrete values that have substrings

    discrete_array[start] up to and including discrete_array[end] has the same value
    """
    i = 0
    for c, g in groupby(discrete_array):
        frames = sum(1 for _ in g)
        yield c, i, i + frames - 1
        i += frames


def check_shape(value, expected_shape, name=''):
    """
    Throws a ValueError if value.shape != expected_shape.
    Args:
        value: Matrix to shape check.
        expected_shape: A tuple or list of integers.
        name: An optional name to add to the exception message.
    """
    if value.shape != tuple(expected_shape):
        raise ValueError('Shape mismatch %s: Expected %s, got %s' %
                         (name, str(expected_shape), str(value.shape)))


def extract_positive_weights(weights, force_numerical_weights=False):
    """
    Extract the indices and values of weights that are positive.
    :param weights: single dimensional weight tensor, or non-floating point array/np.ndarray
    :param force_numerical_weights: whether the weights have to be numerical, or if they can be None on hard weights
    :return: indices, values, number
    """
    # assume floating point type means weights are soft
    if torch.is_tensor(weights) and torch.is_floating_point(weights):
        return _extract_soft_weights(weights)
    indices, weights, N = _extract_hard_weights(weights)
    if force_numerical_weights:
        weights = torch.ones(N, dtype=torch.double)
    return indices, weights, N


def _extract_soft_weights(w):
    neighbours = w > 0
    nw = w[neighbours]
    return neighbours, nw, nw.shape[0]


def _extract_hard_weights(w):
    if torch.is_tensor(w):
        return w, None, w.shape[0]
    if isinstance(w, slice):
        N = w.stop - w.start
        return w, None, N
    else:
        raise RuntimeError("Unhandled weight type {}".format(type(w)))


def _tryint(s):
    try:
        return int(s)
    except:
        return s


def _alphanum_key(s):
    """
    Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    :param s:
    :return:
    """
    return [_tryint(c) for c in re.split('([0-9]+)', s)]


def sort_nicely(l):
    """
    In place sort list in the way that humans expects; from https://nedbatchelder.com/blog/200712/human_sorting.html
    :param l: list
    :return:
    """
    l.sort(key=_alphanum_key)



import torch
import os.path
import logging

logger = logging.getLogger(__name__)


class Serializable():
    """
    Stateful objects that can be saved and loaded
    """

    def save(self, filename):
        """Save any state collected during the run (excludes data given at construction time for simplicity)"""
        state = self.state_dict()
        base_dir = os.path.dirname(filename)
        if not os.path.isdir(base_dir):
            os.makedirs(base_dir, exist_ok=True)
        torch.save(state, filename)
        logger.info("saved checkpoint to %s", filename)

    def load(self, filename) -> bool:
        """Load state previously saved, returning loading success"""
        if not os.path.isfile(filename):
            return False
        try:
            checkpoint = torch.load(filename)
        except RuntimeError as e:
            logger.warning(e)
            checkpoint = torch.load(filename, map_location=torch.device('cpu'))
        try:
            if not self.load_state_dict(checkpoint):
                logger.info("failed to load checkpoint from %s", filename)
                logger.info(checkpoint)
                return False
        except KeyError as e:
            logger.error(e)
            logger.info(checkpoint)
            return False
        logger.info("loaded checkpoint from %s", filename)
        return True

    def state_dict(self) -> dict:
        """State collected during the run"""
        return {}

    def load_state_dict(self, state: dict) -> bool:
        return True



import torch


def identity_select(xu):
    labels = xu.new_empty(xu.shape[0], dtype=torch.long)
    u = xu[:, 0]
    labels[u <= -1] = 0
    labels[(u >= -1) & (u <= 2)] = 1
    labels[u > 2] = 2
    return labels


class RandomSelector:
    def __init__(self, tsf, min_proportion_each_class):
        self.tsf = tsf
        # min proportion of each data point
        self.min_proportion_each_class = min_proportion_each_class
        self.thresholds = None

    def __call__(self, xu):
        N = xu.shape[0]
        labels = xu.new_empty(N, dtype=torch.long)

        with torch.no_grad():
            u = self.tsf(xu)
            u = u.view(-1)
            if self.thresholds is None:
                self._calculate_thresholds(u)

            categories = self._calculate_categories(u)
            for i, cat in enumerate(categories):
                labels[cat] = i

        return labels

    def _calculate_thresholds(self, u):
        N = u.shape[0]
        rr = (torch.min(u), torch.max(u))

        bad_proportion = True
        while bad_proportion:
            bad_proportion = False
            thresholds = torch.rand(2, dtype=u.dtype, device=u.device) * (rr[1] - rr[0]) + rr[0]
            self.thresholds = torch.sort(thresholds)[0]

            categories = self._calculate_categories(u)
            for i, cat in enumerate(categories):
                n = torch.sum(cat)
                # require a minimum proportion of each data point
                if n < N * self.min_proportion_each_class:
                    bad_proportion = True
                    break

    def _calculate_categories(self, u):
        categories = [u <= self.thresholds[0], (u >= self.thresholds[0]) & (u <= self.thresholds[1]),
                      u > self.thresholds[1]]
        return categories



import torch

# 1D input, 1D output
simplest = torch.tensor([[1, 2], [-1, 0], [1, 2]], dtype=torch.double).unsqueeze(2)

# 2D input
multi_input = torch.tensor([[1, 1, 2], [-1, -2, 0], [1, 0.5, 2]], dtype=torch.double).unsqueeze(2)



import abc
import copy
import logging
import typing
from typing import Iterable

import torch
from arm_pytorch_utilities import load_data
from arm_pytorch_utilities import tensor_utils

logger = logging.getLogger(__name__)


class Transformer(abc.ABC):
    """Like sklearn's Transformers
    Different from pytorch's transforms as those are typically applied as you retrieve the data (usually images)
    """

    def __init__(self):
        # allow only fit once; for refit, use a different instance of the preprocessor
        # usually refits are mistakes
        self.fitted = False

    def fit(self, XU, Y, labels=None):
        if self.fitted:
            logger.warning("Ignoring attempt to refit preprocessor")
            return
        self._fit_impl(XU, Y, labels)
        self.fitted = True

    def transform(self, XU, Y, labels=None):
        # apply transformation and return transformed copies of data
        return self.transform_x(XU), self.transform_y(Y), labels

    def update_data_config(self, config: load_data.DataConfig):
        """Change the dimensions of the data's configuration to match what the preprocessing will do"""
        pass

    @abc.abstractmethod
    def transform_x(self, XU):
        """Apply transformation on XU"""

    @abc.abstractmethod
    def transform_y(self, Y):
        """Apply transformation on Y"""

    @abc.abstractmethod
    def invert_transform(self, Y, X=None):
        """Invert transformation on Y with potentially information from X (untransformed)"""

    @abc.abstractmethod
    def invert_x(self, X):
        """Invert transformation on X if possible (for transforms that lose information this may be impossible)"""

    @abc.abstractmethod
    def _fit_impl(self, XU, Y, labels):
        """Fit internal state to training set"""


class Compose(Transformer):
    """Compose a list of preprocessors to be applied in order"""

    def __init__(self, transforms: typing.List[Transformer]):
        # order matters here; applied from left to right
        self.transforms = transforms
        super(Compose, self).__init__()

    def transform(self, XU, Y, labels=None):
        # need to override this because held transforms could override these
        for t in self.transforms:
            XU, Y, labels = t.transform(XU, Y, labels)
        return XU, Y, labels

    def _fit_impl(self, XU, Y, labels):
        for t in self.transforms:
            t._fit_impl(XU, Y, labels)
            XU, Y, labels = t.transform(XU, Y, labels)

    def transform_x(self, XU):
        for t in self.transforms:
            XU = t.transform_x(XU)
        return XU

    def transform_y(self, Y):
        for t in self.transforms:
            Y = t.transform_y(Y)
        return Y

    def invert_transform(self, Y, X=None):
        # need to also reverse order
        for t in reversed(self.transforms):
            Y = t.invert_transform(Y, X)
        return Y

    def invert_x(self, X):
        # need to also reverse order
        for t in reversed(self.transforms):
            X = t.invert_x(X)
        return X

    def update_data_config(self, config: load_data.DataConfig):
        for t in self.transforms:
            t.update_data_config(config)


class NoTransform(Transformer):
    def transform_x(self, XU):
        return XU

    def transform_y(self, Y):
        return Y

    def invert_transform(self, Y, X=None):
        return Y

    def invert_x(self, X):
        return X

    def _fit_impl(self, XU, Y, labels):
        # optionally do some fitting on the training set
        pass


class PolynomialState(NoTransform):
    def __init__(self, order=2):
        super().__init__()
        from sklearn.preprocessing import PolynomialFeatures
        self.poly = PolynomialFeatures(order, include_bias=False)

    def _fit_impl(self, XU, Y, labels):
        self.poly.fit(XU.numpy())

    def transform_x(self, XU):
        p = self.poly.transform(XU.numpy())
        return torch.from_numpy(p)

    def invert_x(self, X):
        raise NotImplemented


class SklearnTransformer(Transformer):
    def __init__(self, method, methodY=None):
        super().__init__()
        self.method = method
        self.methodY = methodY or copy.deepcopy(method)

    def _fit_impl(self, XU, Y, labels):
        self.method.fit(XU)
        self.methodY.fit(Y)

    def transform_x(self, XU):
        x = self.method.transform(XU)
        return torch.from_numpy(x)

    def transform_y(self, Y):
        y = self.methodY.transform(Y)
        return torch.from_numpy(y)

    def invert_transform(self, Y, X=None):
        return torch.from_numpy(self.methodY.inverse_transform(Y))

    def invert_x(self, X):
        return torch.from_numpy(self.method.inverse_transform(X))


class SingleTransformer:
    """Transformer like Sklearn's transformers; pytorch version to pass through differentiation"""

    @abc.abstractmethod
    def fit(self, X):
        """Fit to data"""

    @abc.abstractmethod
    def transform(self, X):
        """Transform data based on previous fit"""

    @abc.abstractmethod
    def inverse_transform(self, X):
        """The inverse transformation"""

    def data_dim_change(self):
        """How this transform changed the dimension of the data.

        Also gives enough info to tell if it changed nx or nu

        :return: (change in dimensions, smallest index of dimension created/removed)
        """
        return 0, None


class NullSingleTransformer(SingleTransformer):
    def fit(self, X):
        pass

    def transform(self, X):
        return X

    def inverse_transform(self, X):
        return X


class SelectTransformer(SingleTransformer):
    """Select some dimensions of incoming data; do nothing on inverse"""

    def __init__(self, cols):
        self.cols = cols

    def fit(self, X):
        tensor_utils.ensure_tensor(X.device, torch.long, self.cols)

    def transform(self, X):
        return X[:, self.cols]

    def inverse_transform(self, X):
        return X


class PytorchTransformer(Transformer):
    def __init__(self, method: SingleTransformer, methodY=None):
        super().__init__()
        self.method = method
        self.methodY = methodY or copy.deepcopy(method)

    def _fit_impl(self, XU, Y, labels):
        self.method.fit(XU)
        self.methodY.fit(Y)

    def transform_x(self, XU):
        return self.method.transform(XU)

    def transform_y(self, Y):
        return self.methodY.transform(Y)

    def invert_transform(self, Y, X=None):
        return self.methodY.inverse_transform(Y)

    def invert_x(self, X):
        return self.method.inverse_transform(X)

    def update_data_config(self, config: load_data.DataConfig):
        change, loc = self.method.data_dim_change()
        if loc is not None:
            if loc < config.nx:
                config.nx += change
            else:
                config.nu += change
            if config.n_input:
                config.n_input += change
        change, _ = self.methodY.data_dim_change()
        config.ny += change


class MinMaxScaler(SingleTransformer):
    def __init__(self, feature_range=(0, 1), dims_share_scale: Iterable[typing.List[int]] = (), **kwargs):
        super().__init__(**kwargs)
        self.feature_range = feature_range
        self.dims_share_scale = dims_share_scale
        self._scale, self._min = None, None

    def fit(self, X):
        # TODO assume we have no nans
        self._fit_with_low_high(torch.min(X, dim=0)[0], torch.max(X, dim=0)[0])

    def _fit_with_low_high(self, low, high):
        feature_range = self.feature_range
        nx = low.shape[0]
        # handle per-dimensional feature range
        if type(feature_range[0]) in (list, tuple):
            feature_range = torch.tensor(feature_range)
        if tensor_utils.is_tensor_like(feature_range):
            # should be 2 x nx; could have fewer columns; would fill with (0,1)
            assert len(feature_range) == 2
            assert feature_range.shape[0] == 2
            f = torch.zeros((2, nx), dtype=low.dtype, device=low.device)
            f[1] = 1
            f[:, :feature_range.shape[1]] = feature_range.to(dtype=low.dtype, device=low.device)
            feature_range = f

        data_range = high - low
        # handle zeros/no variation in that dimension
        data_range[data_range == 0.] = 1.
        self._scale = ((feature_range[1] - feature_range[0]) / data_range)
        for shared_groups in self.dims_share_scale:
            self._scale[shared_groups] = self._scale[shared_groups].mean()
            # note that afterward the shared groups will not have max = feature_range[1]
            # but will still have min = feature_range[0]
        self._min = feature_range[0] - low * self._scale

    def transform(self, X):
        return (X * self._scale) + self._min

    def inverse_transform(self, X):
        return (X - self._min) / self._scale


class RobustMinMaxScaler(MinMaxScaler):
    """Like min-max, but ignore outliers"""

    def __init__(self, percentile=0.975, **kwargs):
        super().__init__(**kwargs)
        self.percentile = percentile

    def fit(self, X):
        low = torch.kthvalue(X, max(1, int((1 - self.percentile) * X.shape[0])), dim=0)[0]
        high = torch.kthvalue(X, int(self.percentile * X.shape[0]), dim=0)[0]
        self._fit_with_low_high(low, high)


class StandardScaler(SingleTransformer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._m = None
        self._s = None

    def fit(self, X):
        self._m = X.mean(0, keepdim=True)
        self._s = X.std(0, unbiased=False, keepdim=True)

    def transform(self, X):
        return (X - self._m) / self._s

    def inverse_transform(self, X):
        return (X * self._s) + self._m


class AngleToCosSinRepresentation(SingleTransformer):
    """
    Continuous representation of an angle which allows better approximation by neural networks
    """

    def __init__(self, angle_index):
        # this transform only handles 1 location; for more locations compose them together
        assert type(angle_index) is int
        self.angle_index = angle_index

    def fit(self, X):
        assert self.angle_index < X.shape[1]

    def transform(self, X):
        s = torch.sin(X[:, self.angle_index]).view(-1, 1)
        c = torch.cos(X[:, self.angle_index]).view(-1, 1)
        return torch.cat((X[:, :self.angle_index], s, c, X[:, self.angle_index + 1:]), dim=1)

    def inverse_transform(self, X):
        s = X[:, self.angle_index]
        c = X[:, self.angle_index + 1]
        theta = torch.atan2(s, c)
        return torch.cat((X[:, :self.angle_index], theta.view(-1, 1), X[:, self.angle_index + 2:]), dim=1)

    def data_dim_change(self):
        return 1, self.angle_index


class DatasetPreprocessor(abc.ABC):
    """Wrapper around Transformer for datasets
    with a transformation fitted on the training set and applied to both training and validation set.
    Assumes the dataset is from a dynamical system (X,U,Y,labels)
    """

    def __init__(self, transformer: Transformer):
        self.tsf = transformer

    def fit(self, dataset):
        if self.tsf.fitted:
            logger.warning("Ignoring attempt to refit preprocessor")
            return
        XU, Y, labels = dataset[:]
        self.tsf.fit(XU, Y, labels)

    def transform(self, dataset):
        if len(dataset):
            XU, Y, labels = self.tsf.transform(*dataset[:])
        else:
            XU, Y, labels = (torch.tensor([], device=series.device) for series in dataset[:])
        return load_data.SimpleDataset(XU, Y, labels)

    def update_data_config(self, config: load_data.DataConfig):
        return self.tsf.update_data_config(config)

    def transform_x(self, XU):
        return self.tsf.transform_x(XU)

    def transform_y(self, Y):
        return self.tsf.transform_y(Y)

    def invert_transform(self, Y, X=None):
        return self.tsf.invert_transform(Y, X)

    def invert_x(self, X):
        return self.tsf.invert_x(X)



import numpy as np
import abc
from arm_pytorch_utilities import serialization


class Controller(serialization.Serializable):
    """
    Controller that gives a command for a given observation (public API is ndarrays)
    Internally may keep state represented as ndarrays or tensors
    """

    def __init__(self, compare_to_goal=np.subtract):
        """
        :param compare_to_goal: function (state, goal) -> diff batched difference
        """
        self.goal = None
        self.compare_to_goal = compare_to_goal

    def state_dict(self) -> dict:
        return {'goal': self.goal}

    def load_state_dict(self, state: dict) -> bool:
        self.goal = state['goal']
        return True

    def reset(self):
        """Clear any controller state to be reused in another trial"""

    def get_goal(self):
        return self.goal

    def set_goal(self, goal):
        self.goal = goal

    @abc.abstractmethod
    def command(self, obs, info=None):
        """Given current observation and misc info, command an action"""

    def get_rollouts(self, obs):
        """Return what the predicted states for the selected action sequence is applied on obs"""
        return None



import math

import torch
from arm_pytorch_utilities import linalg


def clip(a, min_val, max_val):
    """Vectorized torch.clamp (supports tensors for min_val and max_val)"""
    return torch.max(torch.min(a, max_val), min_val)


def replace_nan_and_inf(a, replacement=0):
    """Replaces nan,inf,-inf values with replacement value in place"""
    a[torch.isnan(a)] = replacement
    a[a == float('inf')] = replacement
    a[a == -float('inf')] = replacement
    return a


def get_bounds(u_min=None, u_max=None):
    """Convenience for ensuring a bound if given is two sided"""
    # make sure if any of them is specified, both are specified
    if u_max is not None and u_min is None:
        u_min = -u_max
    if u_min is not None and u_max is None:
        u_max = -u_min
    return u_min, u_max


def rotate_wrt_origin(xy, theta):
    return (xy[0] * math.cos(theta) - xy[1] * math.sin(theta),
            xy[0] * math.sin(theta) + xy[1] * math.cos(theta))


def batch_rotate_wrt_origin(xy, theta):
    # create rotation matrices
    c = torch.cos(theta).view(-1, 1)
    s = torch.sin(theta).view(-1, 1)
    top = torch.cat((c, s), dim=1)
    bot = torch.cat((-s, c), dim=1)
    R = torch.cat((top.unsqueeze(2), bot.unsqueeze(2)), dim=2)
    return linalg.batch_batch_product(xy, R)


def angle_between(u: torch.tensor, v: torch.tensor):
    """Angle between 2 n-dimensional vectors

    :param u N x n tensor
    :param v M x n tensor
    :return N x M angle in radians between each tensor
    """
    u_n = u / u.norm(dim=1, keepdim=True)
    v_n = v / v.norm(dim=1, keepdim=True)
    c = clip(u_n @ v_n.transpose(0, 1),
             torch.tensor(-1, device=u.device, dtype=u.dtype), torch.tensor(1, device=u.device, dtype=u.dtype))
    return torch.acos(c)


def angular_diff(a, b):
    """Angle difference from b to a (a - b)"""
    d = a - b
    if d > math.pi:
        d -= 2 * math.pi
    elif d < -math.pi:
        d += 2 * math.pi
    return d


def angular_diff_batch(a, b):
    """Angle difference from b to a (a - b)"""
    d = a - b
    d[d > math.pi] -= 2 * math.pi
    d[d < -math.pi] += 2 * math.pi
    return d


def angle_normalize(a):
    """Wrap angle between (-pi,pi)"""
    return ((a + math.pi) % (2 * math.pi)) - math.pi


def cos_sim_pairwise(x1, x2=None, eps=1e-8):
    """Pairwise cosine similarity between 2 sets of vectors

    From https://github.com/pytorch/pytorch/issues/11202
    :param x1: (N, nx) N tensors of nx dimension each
    :param x2: (M, nx) M tensors of nx dimension each, equal to x1 if not specified
    :param eps: threshold above 0 to reduce dividing by 0
    :return: (N, M) pairwise cosine similarity
    """
    x2 = x1 if x2 is None else x2
    w1 = x1.norm(p=2, dim=1, keepdim=True)
    w2 = w1 if x2 is x1 else x2.norm(p=2, dim=1, keepdim=True)
    return torch.mm(x1, x2.t()) / (w1 * w2.t()).clamp(min=eps)



import torch
import torch.nn as nn
from torch.distributions import Normal, OneHotCategorical
from torch.nn import functional as F


class MixtureDensityNetwork(nn.Module):
    """
    Mixture density network; implementation mostly from https://github.com/tonyduan/mdn.
    Used as the final/output block of a network.
    [ Bishop, 1994 ]
    Parameters
    ----------
    dim_in: int; dimensionality of the covariates
    dim_out: int; dimensionality of the response variable
    n_components: int; number of components in the mixture model
    """

    def __init__(self, dim_in, dim_out, n_components):
        super().__init__()
        self.pi_network = CategoricalNetwork(dim_in, n_components)
        self.normal_network = MixtureDiagNormalNetwork(dim_in, dim_out,
                                                       n_components)

    def forward(self, x):
        return self.pi_network(x), self.normal_network(x)

    @staticmethod
    def loss(pi, normal, y):
        loglik = normal.log_prob(y.unsqueeze(1).expand_as(normal.loc))
        loglik = torch.sum(loglik, dim=2)
        loss = -torch.logsumexp(torch.log(pi.probs) + loglik, dim=1)
        return loss

    @staticmethod
    def sample(pi, normal):
        samples = torch.sum(pi.sample().unsqueeze(2) * normal.sample(), dim=1)
        return samples

    @staticmethod
    def mean(pi, normal):
        means = torch.sum(pi.mean.unsqueeze(2) * normal.mean, dim=1)
        return means

    @staticmethod
    def stddev(pi, normal):
        # assuming independent
        var = pi.mean.unsqueeze(2) ** 2 * normal.variance + pi.variance.unsqueeze(
            2) * normal.mean ** 2 + pi.variance.unsqueeze(2) * normal.variance
        stddev = torch.sqrt(torch.sum(var, dim=1))
        return stddev


class MixtureDiagNormalNetwork(nn.Module):

    def __init__(self, in_dim, out_dim, n_components, hidden_dim=None):
        super().__init__()
        self.n_components = n_components
        if hidden_dim is None:
            hidden_dim = in_dim
        self.network = nn.Sequential(
            nn.Linear(in_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, 2 * out_dim * n_components),
        )

    def forward(self, x):
        params = self.network(x)
        mean, sd = torch.split(params, params.shape[1] // 2, dim=1)
        mean = torch.stack(mean.split(mean.shape[1] // self.n_components, 1))
        sd = torch.stack(sd.split(sd.shape[1] // self.n_components, 1))
        # improve stability by using elu instead of exp
        sd = F.elu(sd) + 1 + 1e-15
        return Normal(mean.transpose(0, 1), sd.transpose(0, 1))


class CategoricalNetwork(nn.Module):

    def __init__(self, in_dim, out_dim, hidden_dim=None):
        super().__init__()
        if hidden_dim is None:
            hidden_dim = in_dim
        self.network = nn.Sequential(
            nn.Linear(in_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, out_dim)
        )

    def forward(self, x):
        params = self.network(x)
        return OneHotCategorical(logits=params)



import torch


class SoftKNN(torch.nn.Module):
    """
    Differentiable relaxation of the k-nearest neighbour operation.
    For a N x n feature input, we output a N x N weight matrix such that the (i,j) element is the weight of the
    jth data point of being one of the k nearest neighbours of the ith data point. Weights are non-negative and
    their upper value range depends on initialization options.
    """

    def __init__(self, min_k=5, activation='linear', uniform_threshold=False, normalization=None):
        """
        :param min_k: Minimum number of neighbours to consider at each point; the behaviour depends on what activation
        is used and whether a uniform threshold is chosen
        :param activation: The function to apply to the shifted distances to get positive weights; currently there is
        'linear' which is ReLU and otherwise we use sigmoid. We recover hard KNN by using the sign activation function
        (1 for distance >= 0 and 0 otherwise).
        :param uniform_threshold: Whether a single distance threshold should be chosen for all data points such that
        at least min_k points is guaranteed to have positive weights for every data point
        :param normalization: None or the p value for the L_p norm, use 1 for sum to 1, use 2 for norm to 1
        """
        super(SoftKNN, self).__init__()
        self.min_neighbours = min_k
        self.activation = activation
        self.normalization = normalization
        self.uniform_threshold = uniform_threshold

    def forward(self, features):
        # calculate the distances of every point to every other point in feature space
        dists = torch.cdist(features, features)

        # select a distance threshold such that any point beyond this is weighted 0
        # basically the max of the kth smallest element inside each row
        dd, ddi = torch.topk(dists, self.min_neighbours + 1, largest=False, sorted=False, dim=1)
        threshold = dd[:, self.min_neighbours]

        if self.uniform_threshold:
            threshold = torch.max(threshold)
            dd = (threshold - dists)
        else:
            dd = (threshold[:, None] - dists)

        # avoid edge case of multiple elements at kth closest distance causing them to become 0
        dd += 1e-10

        # TODO add more activation options
        # apply weighting function to the distances/nodes
        if self.activation == 'linear':
            weights = dd.clamp(min=0)
        else:
            weights = torch.sigmoid(dd * self.activation)

        # normalization
        if self.normalization is not None:
            weights = torch.nn.functional.normalize(weights, dim=1, p=self.normalization)

        return weights



from arm_pytorch_utilities import preprocess
from arm_pytorch_utilities import math_utils
import torch


class Multiplier(preprocess.SingleTransformer):
    def __init__(self, scale):
        self.scale = scale

    def fit(self, X):
        pass

    def transform(self, X):
        return X * self.scale

    def inverse_transform(self, X):
        return X / self.scale


def verify_scaler_tsf(x, y, tsf, scale=1.):
    tsf.fit(x, y)
    nx = x.shape[1]
    ny = y.shape[1]
    xx, yy, _ = tsf.transform(x, y)
    v, _ = xx.min(0)
    assert torch.allclose(v, torch.zeros(nx))
    v, _ = xx.max(0)
    assert torch.allclose(v, torch.ones(nx) * scale)
    v, _ = yy.min(0)
    assert torch.allclose(v, torch.zeros(ny))
    v, _ = yy.max(0)
    assert torch.allclose(v, torch.ones(ny) * scale)
    yyy = tsf.invert_transform(yy, x)
    assert torch.allclose(y, yyy, atol=1e-6)


def test_preprocess_compose():
    N = 100
    nx = 3
    ny = 2
    scale = 5.5
    x = torch.randn((N, nx))
    y = torch.randn((N, ny))
    tsf1 = preprocess.PytorchTransformer(preprocess.MinMaxScaler())
    tsf2 = preprocess.PytorchTransformer(Multiplier(scale))

    combo1 = preprocess.Compose([tsf1, tsf2])
    # multiplier applied afterwards
    verify_scaler_tsf(x, y, combo1, scale)

    combo2 = preprocess.Compose([tsf2, tsf1])
    # multiplier applied before scaler
    verify_scaler_tsf(x, y, combo2)


def test_min_max_scaler():
    N = 100
    nx = 3
    ny = 2
    x = torch.randn((N, nx))
    y = torch.randn((N, ny))
    tsf = preprocess.PytorchTransformer(preprocess.MinMaxScaler())
    verify_scaler_tsf(x, y, tsf)

    tsf = preprocess.PytorchTransformer(preprocess.MinMaxScaler(feature_range=[[0, 0], [2, 3]]))
    tsf.fit(x, y)
    xx, yy, _ = tsf.transform(x, y)
    v, _ = xx.min(0)
    assert torch.allclose(v, torch.zeros(nx))
    v, _ = xx.max(0)
    assert torch.allclose(v, torch.tensor([2., 3., 1.]))
    v, _ = yy.min(0)
    assert torch.allclose(v, torch.zeros(ny))
    v, _ = yy.max(0)
    assert torch.allclose(v, torch.tensor([2., 3.]))
    yyy = tsf.invert_transform(yy, x)
    assert torch.allclose(y, yyy, atol=1e-6)


def test_min_max_shared_scale():
    N = 100
    nx = 3
    x = torch.randn((N, nx))
    x[:, 1] = x[:, 1] * 2 + 5
    tsf = preprocess.MinMaxScaler(dims_share_scale=[[0, 1]])
    tsf.fit(x)
    xx = tsf.transform(x)
    v, _ = xx.min(0)
    assert torch.allclose(v, torch.zeros(nx))
    v, _ = xx.max(0)
    assert torch.allclose(tsf._scale[0], tsf._scale[1])


def try_robust_min_max_scaler():
    N = 100
    nx = 3
    ny = 2
    x = torch.rand((N, nx)) * 3 - 1
    y = torch.rand((N, ny)) * 2 + 2
    tsf = preprocess.PytorchTransformer(preprocess.RobustMinMaxScaler())
    tsf.fit(x, y)
    xx, yy, _ = tsf.transform(x, y)
    print("xx low {} high {}".format(xx.min(dim=0)[0], xx.max(dim=0)[0]))
    print("yy low {} high {}".format(yy.min(dim=0)[0], yy.max(dim=0)[0]))


def test_angle_to_cos_sin():
    N = 100
    nx = 3
    for angle_index in range(nx):
        x = torch.randn((N, nx))
        tsf = preprocess.AngleToCosSinRepresentation(angle_index)
        tsf.fit(x)
        xx = tsf.transform(x)

        # doesn't touch the other data
        assert torch.allclose(x[:, :angle_index], xx[:, :angle_index])
        assert torch.allclose(x[:, angle_index + 1:], xx[:, angle_index + 2:])
        # is sine and cosine of angle
        assert torch.allclose(torch.sin(x[:, angle_index]), xx[:, angle_index])
        assert torch.allclose(torch.cos(x[:, angle_index]), xx[:, angle_index + 1])
        # converts back ok
        xxx = tsf.inverse_transform(xx)
        assert torch.allclose(x[:, :angle_index], xxx[:, :angle_index])
        assert torch.allclose(x[:, angle_index + 1:], xxx[:, angle_index + 1:])
        assert torch.allclose(
            math_utils.angular_diff_batch(x[:, angle_index], xxx[:, angle_index]),
            torch.zeros_like(x[:, angle_index]), atol=1e-6)


def test_select_transform():
    N = 100
    nx = 6
    x = torch.rand((N, nx)) * 3 - 1
    selected_nx = [0, 1, 4, 5]
    tsf = preprocess.SelectTransformer(selected_nx)
    tsf.fit(x)
    xx = tsf.transform(x)

    assert torch.allclose(xx, x[:, selected_nx])


if __name__ == "__main__":
    test_min_max_scaler()
    test_min_max_shared_scale()
    test_preprocess_compose()
    test_angle_to_cos_sin()
    try_robust_min_max_scaler()
    test_select_transform()



import math

import torch
from arm_pytorch_utilities import math_utils


def test_angle_normalize():
    N = 100
    theta = torch.randn(N) * 5
    t1 = math_utils.angle_normalize(theta)
    assert torch.allclose(torch.cos(theta), torch.cos(t1), atol=1e-6)
    assert torch.allclose(torch.sin(theta), torch.sin(t1), atol=1e-6)


def test_batch_angle_rotate():
    N = 100
    xy = torch.tensor([1., 0]).repeat(N, 1)
    theta = (torch.rand(N) - 0.5) * 2 * math.pi

    xyr = math_utils.batch_rotate_wrt_origin(xy, theta)
    for i in range(N):
        r = math_utils.rotate_wrt_origin(xy[i], theta[i])
        assert torch.allclose(torch.tensor(r), xyr[i])


def test_cos_sim_pairwise():
    N = 100
    M = 30
    nx = 5
    x1 = torch.rand((N, nx))
    x2 = torch.rand((M, nx))
    C = math_utils.cos_sim_pairwise(x1, x2)
    assert C.shape == (N, M)
    for m in range(M):
        c = torch.cosine_similarity(x1, x2[m].view(1, -1))
        assert torch.allclose(c, C[:, m])


def test_angle_between():
    u = torch.tensor([[1., 0, 0]])
    v = torch.tensor([[0., 1, 0], [1, 0, 0]])
    assert torch.allclose(math_utils.angle_between(u, v), torch.tensor([[math.pi / 2, 0]]))

    u = torch.tensor([[1., 0, 0], [-1, 0, 0]])
    res = math_utils.angle_between(u, v)
    assert res.shape == (2, 2)
    assert torch.allclose(res, torch.tensor([[math.pi / 2, 0], [math.pi / 2, math.pi]]))


if __name__ == "__main__":
    test_angle_normalize()
    test_batch_angle_rotate()
    test_cos_sim_pairwise()
    test_angle_between()



import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
import torch
from arm_pytorch_utilities import array_utils
from arm_pytorch_utilities.model.mdn import MixtureDensityNetwork
from matplotlib.patches import Ellipse


def clear_ax_content(ax):
    """Clear a matplotlib axis's content without changing its limits or labels"""
    for artist in ax.lines + ax.collections + ax.patches:
        artist.remove()


def confidence_ellipse(center, cov, ax, n_std=3.0, facecolor='none', **kwargs):
    """
    Create a plot of the covariance confidence ellipse around center with covariance cov

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.

    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.

    Returns
    -------
    matplotlib.patches.Ellipse

    Other parameters
    ----------------
    kwargs : `~matplotlib.patches.Patch` properties
    """
    pearson = cov[0, 1] / np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0),
                      width=ell_radius_x * 2,
                      height=ell_radius_y * 2,
                      facecolor=facecolor,
                      **kwargs)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(center[0], center[1])

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)


def highlight_value_ranges(discrete_array, color_map='rgbcmyk', ymin=0., ymax=1., ax=None, x_values=None):
    """Highlight the background of current figure with a label array; a value of 0 is left blank"""
    for value, start, end in array_utils.discrete_array_to_value_ranges(discrete_array):
        if value == 0:
            continue
        # use the current axis if one is not given (global call on plt module will use current axis)
        if ax is None:
            ax = plt
        if x_values is not None:
            start, end = x_values[start], x_values[end]
        ax.axvspan(start, end, facecolor=color_map[value % len(color_map)], alpha=0.3, ymin=ymin, ymax=ymax)


def plot_state(X, labels, axis_name, title):
    state_dim = X.shape[1]
    assert state_dim == len(axis_name)

    fig, axes = plt.subplots(1, state_dim, figsize=(18, 5))
    for i in range(state_dim):
        axes[i].set_xlabel(axis_name[i])
        axes[i].plot(X[:, i].numpy())
        highlight_value_ranges(labels, ax=axes[i], color_map='rr')
    fig.suptitle(title)


def plot_mdn_prediction(learned_model, X, Y, labels, axis_name, title, output_offset=2, plot_states=False,
                        sample=0):
    N = X.shape[0]
    # freeze model
    for param in learned_model.parameters():
        param.requires_grad = False

    if plot_states:
        plot_state(X, labels, axis_name, title)

    # plot output/prediction (differences)
    output_name = axis_name[output_offset:]
    output_dim = Y.shape[1]
    f2, a2 = plt.subplots(1, output_dim, figsize=(18, 5))

    pi, normal = learned_model(X)
    if sample:
        Yhat = MixtureDensityNetwork.sample(pi, normal)
        # sample multiple times
        for _ in range(1, sample):
            Yhat = torch.cat((Yhat, MixtureDensityNetwork.sample(pi, normal)), dim=0)
    else:
        Yhat = MixtureDensityNetwork.mean(pi, normal)
        stddev = MixtureDensityNetwork.stddev(pi, normal)
        stddev = stddev.numpy()
    Yhat = Yhat.numpy()

    posterior = pi.probs
    modes = np.argmax(posterior, axis=1)

    frames = np.arange(N)
    if sample:
        frames = np.tile(frames, sample)

    for i in range(output_dim):
        j = i
        a2[i].set_xlabel(output_name[i])
        a2[i].plot(Y[:, j])
        if sample:
            a2[i].scatter(frames, Yhat[:, j], alpha=0.4, color='orange')
        else:
            mean = Yhat[:, j]
            std = stddev[:, j]
            a2[i].plot(mean)
            a2[i].fill_between(frames, mean - std, mean + std, color='orange', alpha=0.2)

        highlight_value_ranges(modes, ax=a2[i], ymin=0.5)
        highlight_value_ranges(labels, ax=a2[i], color_map='rr', ymax=0.5)
    f2.suptitle(title)

    plt.figure()
    components = posterior.shape[1]
    for i in range(components):
        plt.plot(posterior[:, i])
    highlight_value_ranges(modes, ymin=0.5)
    highlight_value_ranges(labels, color_map='rr', ymax=0.5)
    plt.title('{} component posterior'.format(title))


def plot_prediction(learned_model, X, Y, labels, axis_name, title, output_offset=2, plot_states=False, sample=None):
    # ignore sample, just there to ensure consistent API with probabilistic models
    # freeze model
    for param in learned_model.parameters():
        param.requires_grad = False

    if plot_states:
        plot_state(X, labels, axis_name, title)

    # plot output/prediction (differences)
    output_name = axis_name[output_offset:]
    output_dim = Y.shape[1]
    f2, a2 = plt.subplots(1, output_dim, figsize=(18, 5))

    Yhat = learned_model(X)
    Yhat = Yhat.numpy()

    for i in range(output_dim):
        j = i
        a2[i].set_xlabel(output_name[i])
        a2[i].plot(Y[:, j])

        mean = Yhat[:, j]
        a2[i].plot(mean)
        highlight_value_ranges(labels, ax=a2[i], color_map='rr', ymax=0.5)
    f2.legend('')

    f2.suptitle(title)


def cumulative_dist(series, labels, xlabel, indicators=None):
    """
    Plot the empirical CDF of some series

    :param series: Iterable of 1D tensors (or a 2D tensor with each row being a series)
    :param labels: Names for each of the series (iterable of strings)
    :param xlabel: Name for the value the series are over
    :param indicators: Indicator variables for each data point in each series to indicate a special event
    :return: figure
    """
    assert len(series) is len(labels)
    f = plt.figure()
    for i in range(len(series)):
        x = series[i]
        label = labels[i]
        assert len(x.shape) == 1
        x = torch.cat((torch.zeros(1, dtype=x.dtype, device=x.device), x))
        N = x.shape[0]
        cumulative_ratio = np.linspace(0, 1, N)
        x, sorted_indices = torch.sort(x)
        p = plt.step(x.cpu().numpy(), cumulative_ratio, where='post', label=label)
        if indicators is not None and indicators[i] is not None:
            ind = torch.cat((torch.zeros(1, dtype=indicators[i].dtype, device=indicators[i].device), indicators[i]))
            ind = ind[sorted_indices]
            plt.plot(x[ind == 1].cpu().numpy(), cumulative_ratio[ind == 1], 'o', color=p[0].get_color(), alpha=0.5)
    plt.xlabel(xlabel)
    plt.ylabel('cumulative ratio')
    plt.ylim(0, 1)
    plt.legend()
    return f



import torch
from arm_pytorch_utilities import grad




def test_jacobian():
    def f1(x):
        return torch.cat((x ** 2, (2 * x[:, 0] + torch.log(x[:, 1])).view(-1, 1)), dim=1)

    nx = 2
    nout_f1 = nx + 1
    x = torch.rand(nx)
    j1 = grad.jacobian(f1, x)
    j2 = grad.jacobian(f1, x.view(1, -1))
    assert j1.shape == (nout_f1, nx)
    assert torch.allclose(j1, j2)
    assert torch.allclose(j1[:, 0], torch.tensor([2 * x[0], 0, 2.]))
    assert torch.allclose(j1[:, 1], torch.tensor([0., 2 * x[1], 1. / x[1]]))

    nx = 5
    nout_f2 = 3
    net = torch.nn.Linear(nx, nout_f2, bias=False)

    def f2(x):
        return net(x)

    x = torch.randn(1, nx)
    # test that it doesn't matter if other operations are ran on the network before calculating jacobian
    f2(x)
    j3 = grad.jacobian(f2, x)
    assert j3.shape == (nout_f2, nx)
    assert torch.allclose(j3, net.weight)

    # can even have existing gradients calculated from some other backward branch
    y = f2(x)
    y.sum().backward()
    j3 = grad.jacobian(f2, x)
    assert torch.allclose(j3, net.weight)


def test_batch_jacobian():
    # test batch jacobian
    N = 10
    nx = 5
    nout_f2 = 3
    net = torch.nn.Linear(nx, nout_f2, bias=False)

    x = torch.rand((N, nx))
    jall = grad.batch_jacobian(net, x)
    assert jall.shape == (N, nout_f2, nx)
    for i in range(N):
        j = grad.jacobian(net, x[i])
        assert torch.allclose(jall[i], j)


if __name__ == "__main__":
    # test_jacobian()
    test_batch_jacobian()



"""from https://github.com/michaelrzhang/lookahead
usage

optimizer = # {any optimizer} e.g. torch.optim.Adam
if args.lookahead:
    optimizer = Lookahead(optimizer, la_steps=args.la_steps, la_alpha=args.la_alpha)
"""
import os
from collections import defaultdict

import torch
from torch.optim.optimizer import Optimizer


def get_device():
    if 'USE_CPU' in os.environ and os.environ['USE_CPU']:
        return torch.device('cpu')
    return torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


class Lookahead(Optimizer):
    r"""PyTorch implementation of the lookahead wrapper.
    Lookahead Optimizer: https://arxiv.org/abs/1907.08610
    """

    def __init__(self, optimizer, la_steps=5, la_alpha=0.8, pullback_momentum="none"):
        """optimizer: inner optimizer
        la_steps (int): number of lookahead steps
        la_alpha (float): linear interpolation factor. 1.0 recovers the inner optimizer.
        pullback_momentum (str): change to inner optimizer momentum on interpolation update
        """
        self.optimizer = optimizer
        self._la_step = 0  # counter for inner optimizer
        self.la_alpha = la_alpha
        self._total_la_steps = la_steps
        pullback_momentum = pullback_momentum.lower()
        assert pullback_momentum in ["reset", "pullback", "none"]
        self.pullback_momentum = pullback_momentum

        self.state = defaultdict(dict)

        # Cache the current optimizer parameters
        for group in optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                param_state['cached_params'] = torch.zeros_like(p.data)
                param_state['cached_params'].copy_(p.data)
                if self.pullback_momentum == "pullback":
                    param_state['cached_mom'] = torch.zeros_like(p.data)

    def __getstate__(self):
        return {
            'state': self.state,
            'optimizer': self.optimizer,
            'la_alpha': self.la_alpha,
            '_la_step': self._la_step,
            '_total_la_steps': self._total_la_steps,
            'pullback_momentum': self.pullback_momentum
        }

    def zero_grad(self):
        self.optimizer.zero_grad()

    def get_la_step(self):
        return self._la_step

    def state_dict(self):
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict)

    def _backup_and_load_cache(self):
        """Useful for performing evaluation on the slow weights (which typically generalize better)
        """
        for group in self.optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                param_state['backup_params'] = torch.zeros_like(p.data)
                param_state['backup_params'].copy_(p.data)
                p.data.copy_(param_state['cached_params'])

    def _clear_and_load_backup(self):
        for group in self.optimizer.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                p.data.copy_(param_state['backup_params'])
                del param_state['backup_params']

    @property
    def param_groups(self):
        return self.optimizer.param_groups

    def step(self, closure=None):
        """Performs a single Lookahead optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = self.optimizer.step(closure)
        self._la_step += 1

        if self._la_step >= self._total_la_steps:
            self._la_step = 0
            # Lookahead and cache the current optimizer parameters
            for group in self.optimizer.param_groups:
                for p in group['params']:
                    param_state = self.state[p]
                    p.data.mul_(self.la_alpha).add_(1.0 - self.la_alpha, param_state['cached_params'])  # crucial line
                    param_state['cached_params'].copy_(p.data)
                    if self.pullback_momentum == "pullback":
                        internal_momentum = self.optimizer.state[p]["momentum_buffer"]
                        self.optimizer.state[p]["momentum_buffer"] = internal_momentum.mul_(self.la_alpha).add_(
                            1.0 - self.la_alpha, param_state["cached_mom"])
                        param_state["cached_mom"] = self.optimizer.state[p]["momentum_buffer"]
                    elif self.pullback_momentum == "reset":
                        self.optimizer.state[p]["momentum_buffer"] = torch.zeros_like(p.data)

        return loss



""" GMM from GPS. """
import logging

import numpy as np
import scipy.linalg

LOGGER = logging.getLogger(__name__)


def logsum(vec, axis=0, keepdims=True):
    maxv = np.max(vec, axis=axis, keepdims=keepdims)
    maxv[maxv == -float('inf')] = 0
    return np.log(np.sum(np.exp(vec - maxv), axis=axis, keepdims=keepdims)) + maxv


class GMM(object):
    """ Gaussian Mixture Model. """

    def __init__(self, init_sequential=False, eigreg=False, warmstart=True):
        self.init_sequential = init_sequential
        self.eigreg = eigreg
        self.warmstart = warmstart
        self.sigma = None

    def inference(self, pts):
        """
        Evaluate dynamics prior.
        Args:
            pts: A N x D array of points.
        """
        # Compute posterior cluster weights.
        logwts = self.clusterwts(pts)

        # Compute posterior mean and covariance.
        mu0, Phi = self.moments(logwts)

        # Set hyperparameters.
        m = self.N
        n0 = m - 2 - mu0.shape[0]

        # Normalize.
        m = float(m) / self.N
        n0 = float(n0) / self.N
        return mu0, Phi, m, n0

    def estep(self, data):
        """
        Compute log observation probabilities under GMM.
        Args:
            data: A N x D array of points.
        Returns:
            logobs: A N x K array of log probabilities (for each point
                on each cluster).
        """
        # Constants.
        K = self.sigma.shape[0]
        Di = data.shape[1]
        N = data.shape[0]

        # Compute probabilities.
        data = data.T
        mu = self.mu[:, 0:Di].T
        mu_expand = np.expand_dims(np.expand_dims(mu, axis=1), axis=1)
        assert mu_expand.shape == (Di, 1, 1, K)
        # Calculate for each point distance to each cluster.
        data_expand = np.tile(data, [K, 1, 1, 1]).transpose([2, 3, 1, 0])
        diff = data_expand - np.tile(mu_expand, [1, N, 1, 1])
        assert diff.shape == (Di, N, 1, K)
        Pdiff = np.zeros_like(diff)
        cconst = np.zeros((1, 1, 1, K))

        for i in range(K):
            U = scipy.linalg.cholesky(self.sigma[i, :Di, :Di],
                                      check_finite=False)
            Pdiff[:, :, 0, i] = scipy.linalg.solve_triangular(
                U, scipy.linalg.solve_triangular(
                    U.T, diff[:, :, 0, i], lower=True, check_finite=False
                ), check_finite=False
            )
            cconst[0, 0, 0, i] = -np.sum(np.log(np.diag(U))) - 0.5 * Di * \
                                 np.log(2 * np.pi)

        logobs = -0.5 * np.sum(diff * Pdiff, axis=0, keepdims=True) + cconst
        assert logobs.shape == (1, N, 1, K)
        logobs = logobs[0, :, 0, :] + self.logmass.T
        return logobs

    def moments(self, logwts):
        """
        Compute the moments of the cluster mixture with logwts.
        Args:
            logwts: A K x 1 array of log cluster probabilities.
        Returns:
            mu: A (D,) mean vector.
            sigma: A D x D covariance matrix.
        """
        # Exponentiate.
        wts = np.exp(logwts)

        # Compute overall mean.
        mu = np.sum(self.mu * wts, axis=0)

        # Compute overall covariance.
        # For some reason this version works way better than the "right"
        # one... could we be computing xxt wrong?
        diff = self.mu - np.expand_dims(mu, axis=0)
        diff_expand = np.expand_dims(diff, axis=1) * \
                      np.expand_dims(diff, axis=2)
        wts_expand = np.expand_dims(wts, axis=2)
        sigma = np.sum((self.sigma + diff_expand) * wts_expand, axis=0)
        return mu, sigma

    def clusterwts(self, data):
        """
        Compute cluster weights for specified points under GMM.
        Args:
            data: An N x D array of points
        Returns:
            A K x 1 array of average cluster log probabilities.
        """
        # Compute probability of each point under each cluster.
        logobs = self.estep(data)

        # Renormalize to get cluster weights.
        logwts = logobs - logsum(logobs, axis=1)

        # Average the cluster probabilities.
        logwts = logsum(logwts, axis=0) - np.log(data.shape[0])
        return logwts.T

    def update(self, data, K, min_iter=0, max_iterations=100):
        """
        Run EM to update clusters.
        Args:
            data: An N x D data matrix, where N = number of data points.
            K: Number of clusters to use.
        """
        # Constants.
        N = data.shape[0]
        Do = data.shape[1]

        LOGGER.debug('Fitting GMM with %d clusters on %d points', K, N)

        if (not self.warmstart or self.sigma is None or
                K != self.sigma.shape[0]):
            # Initialization.
            LOGGER.debug('Initializing GMM.')
            self.sigma = np.zeros((K, Do, Do))
            self.mu = np.zeros((K, Do))
            self.logmass = np.log(1.0 / K) * np.ones((K, 1))
            self.mass = (1.0 / K) * np.ones((K, 1))
            self.N = data.shape[0]
            N = self.N

            # Set initial cluster indices.
            if not self.init_sequential:
                cidx = np.random.randint(0, K, size=(1, N))
            else:
                raise NotImplementedError()

            # Initialize.
            for i in range(K):
                cluster_idx = (cidx == i)[0]
                mu = np.mean(data[cluster_idx, :], axis=0)
                diff = (data[cluster_idx, :] - mu).T
                sigma = (1.0 / K) * (diff.dot(diff.T))
                self.mu[i, :] = mu
                self.sigma[i, :, :] = sigma + np.eye(Do) * 2e-6

        prevll = -float('inf')
        for itr in range(max_iterations):
            # E-step: compute cluster probabilities.
            logobs = self.estep(data)

            # Compute log-likelihood.
            ll = np.sum(logsum(logobs, axis=1))
            LOGGER.debug('GMM itr %d/%d. Log likelihood: %f',
                         itr, max_iterations, ll)
            if np.abs(ll - prevll) < 1e-2:
                LOGGER.debug('GMM convergenced on itr=%d/%d',
                             itr, max_iterations)
                break
            prevll = ll

            # Renormalize to get cluster weights.
            logw = logobs - logsum(logobs, axis=1)
            assert logw.shape == (N, K)

            # Renormalize again to get weights for refitting clusters.
            logwn = logw - logsum(logw, axis=0)
            assert logwn.shape == (N, K)
            w = np.exp(logwn)

            # M-step: update clusters.
            # Fit cluster mass.
            self.logmass = logsum(logw, axis=0).T
            self.logmass = self.logmass - logsum(self.logmass, axis=0)
            assert self.logmass.shape == (K, 1)
            self.mass = np.exp(self.logmass)
            # Reboot small clusters.
            w[:, (self.mass < (1.0 / K) * 1e-4)[:, 0]] = 1.0 / N
            # Fit cluster means.
            w_expand = np.expand_dims(w, axis=2)
            data_expand = np.expand_dims(data, axis=1)
            self.mu = np.sum(w_expand * data_expand, axis=0)
            # Fit covariances.
            wdata = data_expand * np.sqrt(w_expand)
            assert wdata.shape == (N, K, Do)
            for i in range(K):
                # Compute weighted outer product.
                XX = wdata[:, i, :].T.dot(wdata[:, i, :])
                mu = self.mu[i, :]
                self.sigma[i, :, :] = XX - np.outer(mu, mu)

                if self.eigreg:  # Use eigenvalue regularization.
                    raise NotImplementedError()
                else:  # Use quick and dirty regularization.
                    sigma = self.sigma[i, :, :]
                    self.sigma[i, :, :] = 0.5 * (sigma + sigma.T) + \
                                          1e-6 * np.eye(Do)



"""Framework for compatible simulations under the /simulation directory"""
import random
import time
import logging
import os
import scipy.io
import pybullet as p
import pybullet_data
import abc

from arm_pytorch_utilities import rand

logger = logging.getLogger(__name__)


class Mode:
    DIRECT = 0
    GUI = 1


class ReturnMeaning:
    SUCCESS = 0
    ERROR = 1
    REJECTED = 2


class _DefaultConfig:
    DATA_DIR = './data'


class Simulation(abc.ABC):
    def __init__(self, save_dir='raw', mode=Mode.GUI, log_video=False, plot=False, save=False,
                 num_frames=300, sim_step_s=1. / 240., config=_DefaultConfig):
        # simulation meta
        self.save_dir = os.path.join(config.DATA_DIR, save_dir)
        self.mode = mode
        self.log_video = log_video

        # simulation config
        self.num_frames = num_frames
        self.sim_step_s = sim_step_s

        # actions to do
        self.plot = plot
        self.save = save

        # per run state variables
        self.randseed = None

    def run(self, randseed=None, run_name=None):
        if randseed is None:
            rand.seed(int(time.time()))
            randseed = random.randint(0, 1000000)
        logger.debug('random seed: %d', randseed)
        self.randseed = randseed
        rand.seed(randseed)

        ret = self._configure_physics_engine()
        if ret is not ReturnMeaning.SUCCESS:
            return ret

        ret = self._setup_experiment()
        if ret is not ReturnMeaning.SUCCESS:
            return ret

        ret = self._init_data()
        if ret is not ReturnMeaning.SUCCESS:
            return ret

        ret = self._run_experiment()
        if ret is not ReturnMeaning.SUCCESS:
            return ret

        # plot data
        if self.plot:
            self._plot_data()

        # save experiment
        if self.save:
            if not os.path.exists(self.save_dir):
                os.makedirs(self.save_dir)
            run_name = run_name if run_name is not None else randseed
            save_to = os.path.join(self.save_dir, '{}.mat'.format(run_name))
            # export in matlab/numpy compatible format
            scipy.io.savemat(save_to, mdict=self._export_data_dict())
            logger.info("Finished saving to {}".format(save_to))

        return ReturnMeaning.SUCCESS

    @abc.abstractmethod
    def _configure_physics_engine(self):
        return ReturnMeaning.SUCCESS

    def _setup_experiment(self):
        return ReturnMeaning.SUCCESS

    def _init_data(self):
        return ReturnMeaning.SUCCESS

    @abc.abstractmethod
    def _run_experiment(self):
        return ReturnMeaning.SUCCESS

    def _plot_data(self):
        pass

    def _export_data_dict(self):
        return {}


class PyBulletSim(Simulation):
    def __init__(self, realtime_simulation=False, **kwargs):
        super(PyBulletSim, self).__init__(**kwargs)
        self.physics_client = None
        self.realtime = realtime_simulation

    def _configure_physics_engine(self):
        mode_dict = {Mode.GUI: p.GUI, Mode.DIRECT: p.DIRECT}

        # if the mode we gave is in the dict then use it, otherwise use the given mode value as is
        mode = mode_dict.get(self.mode) or self.mode

        self.physics_client = p.connect(mode)  # p.GUI for GUI or p.DIRECT for non-graphical version

        if self.log_video:
            p.startStateLogging(p.STATE_LOGGING_VIDEO_MP4, "{}.mp4".format(self.randseed))

        # use data provided by PyBullet
        p.setAdditionalSearchPath(pybullet_data.getDataPath())  # optionally

        if self.realtime:
            p.setRealTimeSimulation(True)
        else:
            p.setTimeStep(self.sim_step_s)

        return ReturnMeaning.SUCCESS

    def run(self, randseed=None, run_name=None):
        # make sure to always disconnect after running
        ret = super(PyBulletSim, self).run(randseed, run_name)
        p.disconnect(self.physics_client)
        return ret



import torch
import numpy as np
from arm_pytorch_utilities import linalg


def assert_same_cov(A, w=None):
    c1 = np.cov(A, rowvar=False, aweights=w)
    c2 = linalg.cov(torch.tensor(A, dtype=torch.float), aweights=w)
    assert np.linalg.norm(c2.numpy() - c1) < 1e-6


def test_cov():
    a = [1, 2, 3, 4]
    assert_same_cov(a)
    A = [[1, 2], [3, 4]]
    assert_same_cov(A)

    assert_same_cov(a, w=[1, 1, 1, 1])
    assert_same_cov(a, w=[2, 0.5, 3, 1])

    assert_same_cov(A, w=[1, 1])
    assert_same_cov(A, w=[2, 0.5])


def test_sqrtm():
    from torch.autograd import gradcheck
    k = torch.randn(20, 10).double()
    # Create a positive definite matrix
    pd_mat = k.t().matmul(k)
    pd_mat.requires_grad = True
    test = gradcheck(linalg.sqrtm, (pd_mat,))
    assert test is True


def test_batch_outer_prodcut():
    n = 5
    N = 10
    u = torch.rand(N, n)
    v = torch.rand(N, n)
    UV = linalg.batch_outer_product(u, v)
    assert UV.shape == (N, n, n)
    for i in range(N):
        uv = torch.ger(u[i], v[i])
        assert torch.allclose(uv, UV[i])


def test_batch_batch_product():
    B = 1000
    ny = 10
    nx = 20
    A = torch.rand(B, ny, nx)
    X = torch.rand(B, nx)

    Y = linalg.batch_batch_product(X, A)
    assert Y.shape == (B, ny)
    for i in range(B):
        y = A[i] @ X[i]
        assert torch.allclose(Y[i], y)


if __name__ == "__main__":
    test_cov()
    test_sqrtm()
    test_batch_outer_prodcut()
    test_batch_batch_product()



import torch


def jacobian(f, x):
    """
    Compute the jacobian of f evaluated at x. If f is a network or has parameters, freezing it
    (set requires_grad = False) improves performance (especially on large networks)

    From https://gist.github.com/sbarratt/37356c46ad1350d4c30aefbd488a4faa
    :param f: f(x) -> y
    :param x: 1 x nx or nx input to evaluate the jacobian at
    :return: df/dx(x) ny x nx jacobian of f evaluated at x
    """
    if x.dim() < 2:
        x = x.view(1, -1)
    y = f(x)
    ny = 1 if len(y.shape) < 2 else y.shape[1]
    x = x.repeat(ny, 1)
    x.requires_grad_(True)
    y = f(x)
    if ny == 1:
        y.backward()
    else:
        y.backward(torch.eye(ny, dtype=y.dtype, device=y.device))
    return x.grad.data


def batch_jacobian(f, x):
    """
    Compute jacobian for a batch of x
    :param f: f(x) -> y (supporting batch input B x <some num> x nx), such as a NN
    :param x: B x nx
    :return: df/dx(x) B x ny x nx jacobian of f evaluated at each x
    """
    y = f(x)
    ny = y.shape[1]

    x = x.unsqueeze(1)  # b, 1, nx
    b = x.shape[0]
    x = x.repeat(1, ny, 1)  # b, ny, nx
    x.requires_grad_(True)
    y = f(x)
    input_val = torch.eye(ny, dtype=y.dtype, device=y.device).repeat(b, 1, 1)
    y.backward(input_val)
    return x.grad.data



""" This file defines the linear Gaussian policy class from GPS. """
import numpy as np
import scipy as sp
from arm_pytorch_utilities.array_utils import check_shape


class LinearGaussianPolicy:
    """
    Time-varying linear Gaussian policy.
    U = K*x + k + noise, where noise ~ N(0, chol_pol_covar)
    """

    def __init__(self, K, k, pol_covar, chol_pol_covar, inv_pol_covar, cache_kldiv_info=False):
        # Assume K has the correct shape, and make sure others match.
        self.T = K.shape[0]
        self.dU = K.shape[1]
        self.dX = K.shape[2]

        check_shape(k, (self.T, self.dU))
        check_shape(pol_covar, (self.T, self.dU, self.dU))
        check_shape(chol_pol_covar, (self.T, self.dU, self.dU))
        check_shape(inv_pol_covar, (self.T, self.dU, self.dU))

        self.K = K
        self.k = k
        self.pol_covar = pol_covar
        self.chol_pol_covar = chol_pol_covar
        self.inv_pol_covar = inv_pol_covar

        # KL Div stuff
        if cache_kldiv_info:
            self.logdet_psig = np.zeros(self.T)
            self.precision = np.zeros((self.T, self.dU, self.dU))
            for t in range(self.T):
                self.logdet_psig[t] = 2 * sum(np.log(np.diag(chol_pol_covar[t])))
                self.precision[t] = sp.linalg.solve_triangular(chol_pol_covar[t],
                                                               sp.linalg.solve_triangular(chol_pol_covar[t].T,
                                                                                          np.eye(self.dU), lower=True,
                                                                                          check_finite=False),
                                                               check_finite=False)

    def act(self, x, obs, t, noise=None, sample=None):
        """
        Return an action for a state.
        Args:
            x: State vector.
            obs: Observation vector.
            t: Time step.
            noise: Action noise. This will be scaled by the variance.
        """
        # import pdb; pdb.set_trace()
        u = self.K[t].dot(x) + self.k[t]
        u += self.chol_pol_covar[t].T.dot(noise)
        return u

    def fold_k(self, noise):
        """
        Fold noise into k.
        Args:
            noise: A T x Du noise vector with mean 0 and variance 1.
        Returns:
            k: A T x dU bias vector.
        """
        k = np.zeros_like(self.k)
        for i in range(self.T):
            scaled_noise = self.chol_pol_covar[i].T.dot(noise[i])
            k[i] = scaled_noise + self.k[i]
        return k

    def nans_like(self):
        """
        Returns:
            A new linear Gaussian policy object with the same dimensions
            but all values filled with NaNs.
        """
        policy = LinearGaussianPolicy(
            np.zeros_like(self.K), np.zeros_like(self.k),
            np.zeros_like(self.pol_covar), np.zeros_like(self.chol_pol_covar),
            np.zeros_like(self.inv_pol_covar)
        )
        policy.K.fill(np.nan)
        policy.k.fill(np.nan)
        policy.pol_covar.fill(np.nan)
        policy.chol_pol_covar.fill(np.nan)
        policy.inv_pol_covar.fill(np.nan)
        return policy



"""Verify that our soft KNN method passes through gradients and is able to learn parameters through it"""
import logging

import torch
from arm_pytorch_utilities import load_data, rand
from arm_pytorch_utilities import softknn
from matplotlib import pyplot as plt

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,
                    format='[%(levelname)s %(asctime)s %(pathname)s:%(lineno)d] %(message)s',
                    datefmt='%m-%d %H:%M:%S')


# generate ground truth transforms and operations
# converts it to a transformed space
class SimpleNet(torch.nn.Module):
    def __init__(self, D_in, H):
        super(SimpleNet, self).__init__()
        self.linear1 = torch.nn.Linear(D_in, H, bias=False)
        self.knn = softknn.SoftKNN(min_k=20)

    def forward(self, x, y):
        """
        In the forward function we accept a Tensor of input data and we must return
        a Tensor of output data. We can use Modules defined in the constructor as
        well as arbitrary operators on Tensors.
        """
        features = self.linear1(x)

        weights = self.knn(features)
        # TODO weighted least squares in x-y space rather than feature space using these weights
        # propagate gradient backwards to transform parameters through these weights
        # TODO remove: for now can just sum up xs inside the neighbourhood

        output = torch.zeros_like(x)
        for i, w in enumerate(weights):
            # nw = w / torch.sum(w)
            # output[i] = torch.matmul(nw, x)
            # can drop out the terms that have 0 weight and seeing if it will affect differentiability
            neighbours = torch.nonzero(w).view(-1)
            nw = w[neighbours]
            nw /= torch.sum(nw)
            output[i] = torch.matmul(nw, x[neighbours])

        return output


def KNN(features, k):
    # features = features.float()
    dist_mat = torch.cdist(features, features)
    # ith row are the k nearest neighbours of ith data point
    dists, Idx = torch.topk(dist_mat, k, largest=False, sorted=False, dim=1)
    return dists, Idx


def test_softknn(debug=False):
    # doesn't always converge in time for all random seed
    seed = 318455
    logger.info('random seed: %d', rand.seed(seed))

    D_in = 3
    D_out = 1

    target_params = torch.rand(D_in, D_out).t()
    # target_params = torch.tensor([[1, -1, 1]], dtype=torch.float )
    target_tsf = torch.nn.Linear(D_in, D_out, bias=False)
    target_tsf.weight.data = target_params
    for param in target_tsf.parameters():
        param.requires_grad = False

    def produce_output(X):
        # get the features
        y = target_tsf(X)
        # cluster in feature space
        dists, Idx = KNN(y, 5)

        # take the sum inside each neighbourhood
        # TODO do a least square fit over X inside each neighbourhood
        features2 = torch.zeros_like(X)
        for i in range(dists.shape[0]):
            # md = max(dists[i])
            # d = md - dists[i]
            # w = d / torch.norm(d)
            features2[i] = torch.mean(X[Idx[i]], 0)
            # features2[i] = torch.matmul(w, X[Idx[i]])

        return features2

    N = 400
    ds = load_data.RandomNumberDataset(produce_output, num=400, input_dim=D_in)
    train_set, validation_set = load_data.split_train_validation(ds)
    train_loader = torch.utils.data.DataLoader(train_set, batch_size=N, shuffle=True)
    val_loader = torch.utils.data.DataLoader(validation_set, batch_size=N, shuffle=False)

    criterion = torch.nn.MSELoss(reduction='sum')

    model = SimpleNet(D_in, D_out)
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-4)

    losses = []
    vlosses = []
    pdist = []
    cosdist = []

    def evaluateLoss(data):
        # target
        x, y = data
        pred = model(x, y)

        loss = criterion(pred, y)
        return loss

    def evaluateValidation():
        with torch.no_grad():
            loss = sum(evaluateLoss(data) for data in val_loader)
            return loss / len(val_loader.dataset)

    # model.linear1.weight.data = target_params.clone()
    for epoch in range(200):
        for i, data in enumerate(train_loader, 0):
            optimizer.zero_grad()

            loss = evaluateLoss(data)
            loss.backward()
            optimizer.step()

            avg_loss = loss.item() / len(data[0])

            losses.append(avg_loss)
            vlosses.append(evaluateValidation())
            pdist.append(torch.norm(model.linear1.weight.data - target_params))
            cosdist.append(torch.nn.functional.cosine_similarity(model.linear1.weight.data, target_params))
            if debug:
                print('[%d, %5d] loss: %.3f' % (epoch + 1, i + 1, avg_loss))

    if debug:
        print('Finished Training')
        print('Target params: {}'.format(target_params))
        print('Learned params:')
        for param in model.parameters():
            print(param)

        print('validation total loss: {:.3f}'.format(evaluateValidation()))

    model.linear1.weight.data = target_params.clone()
    target_loss = evaluateValidation()

    if debug:
        print('validation total loss with target params: {:.3f}'.format(target_loss))

        plt.plot(range(len(losses)), losses)
        plt.plot(range(len(losses)), vlosses)
        plt.plot(range(len(losses)), [target_loss] * len(losses), linestyle='--')
        plt.legend(['training minibatch', 'whole validation', 'validation with target params'])
        plt.xlabel('minibatch')
        plt.ylabel('MSE loss')

        plt.figure()
        plt.plot(range(len(pdist)), pdist)
        plt.xlabel('minibatch')
        plt.ylabel('euclidean distance of model params from target')

        plt.figure()
        plt.plot(range(len(cosdist)), cosdist)
        plt.xlabel('minibatch')
        plt.ylabel('cosine similarity between model params and target')
        plt.show()

    # check that we're close to the actual KNN performance on validation set
    last_few = 5
    loss_tolerance = 0.02
    assert sum(vlosses[-last_few:]) / last_few - target_loss < target_loss * loss_tolerance


if __name__ == "__main__":
    test_softknn(True)



import logging
import os
import itertools

import torch
from arm_pytorch_utilities import array_utils
from arm_pytorch_utilities.optim import Lookahead

logger = logging.getLogger(__name__)


# TODO make subclass of nn.Module?
class LearnableParameterizedModel:
    def __init__(self, save_root_dir, name='', lookahead=False):
        self.save_root_dir = save_root_dir
        self.step = 0
        self.name = name
        params = self.parameters()
        self.optimizer = torch.optim.Adam(params if len(params) else [torch.zeros(1)])
        if lookahead:
            self.optimizer = Lookahead(self.optimizer)

    def modules(self):
        """PyTorch compatible modules that make it convenient to define the other methods"""
        return {}

    def parameters(self):
        """
        :return: Iterable holding this transform's parameters
        """
        return list(itertools.chain.from_iterable(module.parameters() for module in self.modules().values()))

    @property
    def device(self):
        return self.parameters()[0].device

    @property
    def dtype(self):
        return self.parameters()[0].dtype

    def _model_state_dict(self):
        """
        :return: State dictionary of the model to save
        """
        return {key: module.state_dict() for key, module in self.modules().items()}

    def _load_model_state_dict(self, saved_state_dict):
        """
        Load saved state dictionary
        :param saved_state_dict: what _model_state_dict returns
        :return:
        """
        modules = self.modules()
        try:
            for key, module in modules.items():
                module.load_state_dict(saved_state_dict[key])
        except KeyError as e:
            logger.error("Abort loading due to %s", e)
            return False
        return True

    def eval(self):
        for param in self.parameters():
            param.requires_grad = False

    def train(self):
        for param in self.parameters():
            param.requires_grad = True

    def get_last_checkpoint(self, sort_by_time=True):
        """
        Get the last checkpoint for a model matching this one's name (with the highest number of training steps)
        :return: either '' or the filename of the last checkpoint
        """
        base_dir = os.path.join(self.save_root_dir, 'checkpoints')
        # look for files with this base name, what follows should be step.tar
        # (use . after name to prevent matching prefix)
        checkpoints = [filename for filename in os.listdir(base_dir) if filename.startswith(self.name + '.')]
        if not checkpoints:
            return ''
        if sort_by_time:
            checkpoints = [os.path.join(base_dir, filename) for filename in checkpoints]
            checkpoints.sort(key=os.path.getmtime)
            return checkpoints[-1]
        else:
            # order by name
            array_utils.sort_nicely(checkpoints)
            return os.path.join(base_dir, checkpoints[-1])

    def save(self, last=False):
        """
        Save the model to a checkpoint
        :param last: whether it's the last save of the training and to log where we saved to
        :return:
        """
        state = {
            'step': self.step,
            'state_dict': self._model_state_dict(),
            'optimizer': self.optimizer.state_dict()
        }
        base_dir = os.path.join(self.save_root_dir, 'checkpoints')
        if not os.path.isdir(base_dir):
            os.makedirs(base_dir, exist_ok=True)
        full_name = os.path.join(base_dir, '{}.{}.tar'.format(self.name, self.step))
        torch.save(state, full_name)
        if last:
            logger.info("saved checkpoint %s", full_name)

    def load(self, filename):
        """
        Load the model from the filename for a checkpoint
        :param filename:
        :return: whether we successfully loaded the checkpoint or not
        """
        if not os.path.isfile(filename):
            return False
        try:
            checkpoint = torch.load(filename)
        except RuntimeError as e:
            logger.warning(e)
            checkpoint = torch.load(filename, map_location=torch.device('cpu'))
        self.step = checkpoint['step']
        if not self._load_model_state_dict(checkpoint['state_dict']):
            return False
        self.optimizer.load_state_dict(checkpoint['optimizer'])
        logger.info("loaded checkpoint at step %d for %s", self.step, self.name)
        return True



import torch
import torch.nn
from arm_pytorch_utilities.model.mdn import MixtureDensityNetwork
from arm_pytorch_utilities.load_data import DataConfig


def make_fully_connected_layers(input_dim=7, h_units=(32, 32, 32), bias=True, activation_factory=torch.nn.LeakyReLU):
    layers = []
    h_layers = len(h_units)
    for i in range(h_layers):
        in_dim = input_dim if i == 0 else h_units[i - 1]
        out_dim = h_units[i]
        layers.append(torch.nn.Linear(in_dim, out_dim, bias=bias))
        layers.append(activation_factory())
    return layers


def make_mdn_end_block(num_components=4):
    def make_block(fc_output_dim, output_dim):
        return MixtureDensityNetwork(fc_output_dim, output_dim, num_components)

    return make_block


def make_linear_end_block(bias=True, activation=None):
    def make_block(fc_output_dim, output_dim):
        layers = [torch.nn.Linear(fc_output_dim, output_dim, bias=bias)]
        if activation is not None:
            layers.append(activation)
        return torch.nn.Sequential(*layers).double()

    return make_block


def make_sequential_network(config: DataConfig, end_block_factory=make_linear_end_block(), h_units=(32, 32),
                            **kwargs):
    # must have end block (otherwise we would always leave with an activation)
    if end_block_factory is None:
        raise RuntimeError("Need an end block to network")
    # setup input and output sizes based on data
    input_dim = config.input_dim()
    output_dim = config.ny

    layers = make_fully_connected_layers(input_dim=input_dim, h_units=h_units, **kwargs)
    end_block_input = h_units[-1] if len(h_units) else input_dim
    layers.append(end_block_factory(end_block_input, output_dim))

    network = torch.nn.Sequential(*layers).double()
    return network



import abc
import os

import numpy as np
import scipy.io
import torch
import torch.utils.data as data


class DataConfig:
    """
    Data class holding configuration information about a dataset
    """

    def __init__(self, sort_data=True, predict_difference=True, predict_all_dims=True, force_affine=False,
                 expanded_input=False, y_in_x_space=True):
        """

        :param sort_data: Whether the experiments (data between files) are sorted (by random seed id)
        :param predict_difference: Whether the prediction should be the state difference or the next state
        :param predict_all_dims: Whether the prediction should include all state dimensions (some datasets do this regardless)
        :param force_affine: Whether a column of 1s should be added to XU to make the dataset effectively affine
        :param expanded_input: Whether the input has extra dimensions (packed in control dimension)
        :param y_in_x_space: Whether the output is in the same space as state; if not then a lot of assumptions will not hold
        """
        self.sort_data = sort_data
        self.predict_difference = predict_difference
        self.predict_all_dims = predict_all_dims
        self.force_affine = force_affine
        self.expanded_input = expanded_input
        self.y_in_x_space = y_in_x_space
        # unknown quantities until we encounter data (optional)
        self.nx = None
        self.nu = None
        self.ny = None
        # sometimes the full input is larger than xu (such as with expanded input)
        self.n_input = None

    def load_data_info(self, x, u=None, y=None, full_input=None):
        self.nx = x.shape[1]
        if u is not None:
            self.nu = u.shape[1]
        if y is not None:
            self.ny = y.shape[1]
        if full_input is not None:
            self.n_input = full_input.shape[1]
        if self.expanded_input and full_input is None:
            raise RuntimeError("Need to load full input with expanded input")

    def input_dim(self):
        if not self.nx:
            raise RuntimeError("Need to load data info first before asking for input dim")
        if self.n_input:
            return self.n_input
        # else assume we're inputting xu
        ni = self.nx
        if self.nu:
            ni += self.nu
        return ni

    def options(self):
        return self.sort_data, self.predict_difference, self.predict_all_dims, self.force_affine, \
               self.expanded_input, self.y_in_x_space

    def __str__(self):
        return "i{}_o{}_s{}_pd{}_pa{}_a{}_e{}_y{}".format(self.input_dim(), self.ny,
                                                          *(int(config) for config in self.options()))

    def __repr__(self):
        return "DataConfig(sort_data={}, predict_difference={}, predict_all_dims={}, force_affine={}, " \
               "expanded_input={}, y_in_x_space={})".format(*self.options())


class DataLoader(abc.ABC):
    """
    Driver for loading a dataset from file.
    Each dataset should subclass DataLoader and specialize process file raw data to saved content.
    """

    def __init__(self, file_cfg=None, config=DataConfig()):
        if file_cfg is None:
            raise RuntimeError("Incomplete specification of DataLoader")
        self.file_cfg = {key: value for key, value in file_cfg.__dict__.items() if
                         not key.startswith('_') and isinstance(value, str)}
        self.config = config

    @abc.abstractmethod
    def _process_file_raw_data(self, d):
        """
        Turn a file's dictionary content into a data sequence, each element of which has the same number of rows
        :param d: file's dictionary content
        :return: tuple of data sequence
        """

    def load_file(self, full_filename, data):
        raw_data = scipy.io.loadmat(full_filename)
        file_data = self._process_file_raw_data(raw_data)
        if data is None:
            data = list(file_data)
        else:
            for i in range(len(data)):
                data[i] = np.row_stack((data[i], file_data[i]))
        return data

    def load(self, dir, override_config=None):
        if override_config:
            self.config = override_config
        data = None
        full_dir = os.path.join(self.file_cfg['DATA_DIR'], dir)

        if os.path.isfile(full_dir):
            data = self.load_file(full_dir, data)
        else:
            files = os.listdir(full_dir)
            # consistent with the way MATLAB loads files
            if self.config.sort_data:
                files = sorted(files)

            for file in files:
                full_filename = '{}/{}'.format(full_dir, file)
                if os.path.isdir(full_filename):
                    continue
                data = self.load_file(full_filename, data)
        return data


def make_affine(X):
    N = X.shape[0]
    return torch.cat((X, torch.ones((N, 1), dtype=X.dtype)), dim=1)


class RandomNumberDataset(data.Dataset):
    def __init__(self, produce_output, num=1000, low=-1, high=1, input_dim=1):
        r = high - low
        self.x = torch.rand((num, input_dim)) * r / 2 + (low + high) / 2
        self.y = produce_output(self.x)
        super(RandomNumberDataset, self).__init__()

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]

    def get_input(self):
        return self.x

    def get_output(self):
        return self.y


class SimpleXUYDataset(data.Dataset):
    def __init__(self, XU, Y):
        self.XU = XU
        self.Y = Y
        super(SimpleXUYDataset, self).__init__()

    def __len__(self):
        return len(self.XU)

    def __getitem__(self, idx):
        return self.XU[idx], self.Y[idx]


class SimpleDataset(data.Dataset):
    def __init__(self, *sequences):
        self.sequences = sequences

    def __len__(self):
        return len(self.sequences[0])

    def __getitem__(self, idx):
        return tuple(sequence[idx] if sequence is not None else [] for sequence in self.sequences)


class IndexedDataset(SimpleDataset):
    """Same as before, but with last element as the index of the data point for using in Dataloaders"""

    def __getitem__(self, idx):
        return tuple(sequence[idx] if sequence is not None else [] for sequence in self.sequences) + (idx,)


class LoaderXUYDataset(data.Dataset):
    def __init__(self, loader: DataLoader, dirs=('raw',), filter_on_labels=None, max_num=None,
                 config=DataConfig(), device="cpu"):
        if type(dirs) is str:
            dirs = [dirs]
        self.XU = None
        self.Y = None
        self.labels = None
        for dir in dirs:
            XU, Y, labels = loader.load(dir, config)
            if self.XU is None:
                self.XU = XU
                self.Y = Y
                self.labels = labels
            else:
                self.XU = np.row_stack((self.XU, XU))
                self.Y = np.row_stack((self.Y, Y))
                self.labels = np.row_stack((self.labels, labels[:, :self.labels.shape[1]]))
        self._convert_types(device)
        if filter_on_labels:
            self.XU, self.Y, self.labels = filter_on_labels(self.XU, self.Y, self.labels)

        if config.force_affine:
            self.XU = make_affine(self.XU)

        if max_num is not None:
            self.XU = self.XU[:max_num]
            self.Y = self.Y[:max_num]
            self.labels = self.labels[:max_num]

        super().__init__()

    def _convert_types(self, device):
        self.XU = torch.from_numpy(self.XU).to(device=device, dtype=torch.double)
        self.Y = torch.from_numpy(self.Y).to(device=device, dtype=torch.double)
        self.labels = torch.from_numpy(self.labels).to(device=device, dtype=torch.double)

    def __len__(self):
        return self.XU.shape[0]

    def __getitem__(self, idx):
        return self.XU[idx], self.Y[idx], self.labels[idx]


def split_train_validation(dataset, validation_ratio=0.1):
    # consider giving a shuffle (with np.random.shuffle()) option to permute the data before viewing
    offset = int(len(dataset) * (1 - validation_ratio))
    return data.Subset(dataset, range(0, offset)), data.Subset(dataset, range(offset, len(dataset)))


def merge_data_in_dir(config, dir, out_filename, sort=True):
    full_dir = os.path.join(config.DATA_DIR, dir)

    files = os.listdir(full_dir)
    if sort:
        files = sorted(files)

    data = None
    for file in files:
        full_filename = '{}/{}'.format(full_dir, file)
        raw_data = scipy.io.loadmat(full_filename)
        if data is None:
            data = raw_data
        else:
            for key in raw_data.keys():
                data[key] = np.row_stack((data[key], raw_data[key]))
    merged_filename = '{}/{}.mat'.format(config.DATA_DIR, out_filename)
    scipy.io.savemat(merged_filename, data)



"""Trajectory optimization (just LQR from GPS's source code)"""
import numpy as np
import scipy as sp
from numpy.linalg import LinAlgError
import logging

from arm_pytorch_utilities.policy.lin_gauss import LinearGaussianPolicy

LOGGER = logging.getLogger(__name__)

CHECK_FINITE = False  # False is faster


def lu_solve(L, U, A):
    """Solves LUX=A for X"""
    return sp.linalg.solve_triangular(U, sp.linalg.solve_triangular(L, A, lower=True,
                                                                    check_finite=CHECK_FINITE),
                                      check_finite=CHECK_FINITE)


def solve_psd(A, B, reg=0):
    """Solve AX=B via cholesky decomposition (A must be positive semidefinite)"""
    chol = sp.linalg.cholesky(A + reg * np.eye(A.shape[0]))
    return lu_solve(chol.T, chol, B)


def invert_psd(A, reg=0):
    """Invert a PSD matrix via Cholesky + Triangular solves"""
    return solve_psd(A, np.eye(A.shape[0]), reg=reg)


def lqr(cost, lgpolicy, dynamics,
        horizon, T, x, prevx, prevu,
        reg_mu, reg_del, del0, min_mu, discount,
        jacobian=None,
        max_time_varying_horizon=20):
    """
    TODO: Clean up args...

    Plain LQR.
    Returns:
        LinearGaussianPolicy: A new time-varying policy valid for <horizon> timesteps
        reg_mu: Value function regularization (from Tassa synthesis paper, currently unused)
        reg_del:
    """
    dX = prevx.shape[0]
    dU = prevu.shape[0]
    ix = slice(dX)
    iu = slice(dX, dX + dU)
    it = slice(dX + dU)
    ip = slice(dX + dU, dX + dU + dX)

    # Compute forward pass
    cv, Cm, Fd, fc, _, _ = \
        estimate_cost(T, cost, lgpolicy, dynamics, horizon, x, prevx, prevu,
                      max_time_varying_horizon, jacobian=jacobian)

    # Compute optimal policy with short horizon MPC.
    fail = True
    decrease_mu = True
    K = np.zeros((horizon, dU, dX))
    pSig = np.zeros((horizon, dU, dU))
    invPSig = np.zeros((horizon, dU, dU))
    cholPSig = np.zeros((horizon, dU, dU))
    k = np.zeros((horizon, dU))
    while fail:
        Vxx = np.zeros((dX, dX))
        Vx = np.zeros(dX)
        fail = False
        for t in range(horizon - 1, -1, -1):
            F = Fd[t]
            f = fc[t]

            Qtt = Cm[t]
            Qt = cv[t]

            Vxx = Vxx  # + reg_mu*np.eye(dX)  #Note: reg_mu is currently unused
            Qtt = Qtt + F.T.dot(Vxx.dot(F))
            Qt = Qt + F.T.dot(Vx) + F.T.dot(Vxx).dot(f)

            Qtt = 0.5 * (Qtt + Qtt.T)

            try:
                U = sp.linalg.cholesky(Qtt[iu, iu], check_finite=False)
                L = U.T
            except LinAlgError:
                fail = True
                decrease_mu = False
                break
            K[t] = -lu_solve(L, U, Qtt[iu, ix])
            k[t] = -lu_solve(L, U, Qt[iu])
            pSig[t] = lu_solve(L, U, np.eye(dU))
            invPSig[t] = invert_psd(pSig[t])
            cholPSig[t] = sp.linalg.cholesky(pSig[t], check_finite=False)

            # Compute value function.
            Vxx = discount * (Qtt[ix, ix] + Qtt[ix, iu].dot(K[t]))
            Vx = discount * (Qt[ix] + Qtt[ix, iu].dot(k[t]))
            Vxx = 0.5 * (Vxx + Vxx.T)

        # Tassa regularization scheme
        if fail:
            if reg_mu > 1e5:
                raise ValueError("Failed to find SPD solution")
            else:
                reg_del = max(del0, reg_del * del0)
                reg_mu = max(min_mu, reg_mu * reg_del)
                # LOGGER.debug('[LQR reg] Increasing mu -> %f', reg_mu)
        elif decrease_mu:
            reg_del = min(1 / del0, reg_del / del0)
            delmu = reg_del * reg_mu
            if delmu > min_mu:
                reg_mu = delmu
            else:
                reg_mu = min_mu
                # LOGGER.debug('[LQR reg] Decreasing mu -> %f', reg_mu)

    policy = LinearGaussianPolicy(K, k, pSig, cholPSig, invPSig)

    # plot new
    # self.forward(horizon, x, lgpolicy, T, hist_key='new')

    return policy, reg_mu, reg_del


def estimate_cost(cur_timestep, cost, lgpolicy, dynamics, horizon, x0, prevx, prevu, max_time_varying_horizon,
                  jacobian=None, time_varying_dynamics=True):
    """
    Returns cost derivatives and computed dynamics via a forward pass
    """
    # Cost + dynamics estimation

    H = horizon

    N = 1

    cholPSig = lgpolicy.chol_pol_covar
    # PSig = lgpolicy.pol_covar
    K = lgpolicy.K
    k = lgpolicy.k

    dX = K.shape[2]
    dU = K.shape[1]
    dT = dX + dU
    ix = slice(dX)
    iu = slice(dX, dX + dU)

    # Run forward pass

    # Allocate space.
    trajsig = np.zeros((H, dT, dT))
    mu = np.zeros((H, dT))

    trajsig[0, ix, ix] = np.zeros((dX, dX))
    mu[0, ix] = x0

    F = np.zeros((H, dX, dT))
    f = np.zeros((H, dX))
    dynsig = np.zeros((H, dX, dX))

    # HACK
    mu[-1, ix] = prevx
    mu[-1, iu] = prevu
    # Perform forward pass.
    for t in range(H):
        PSig = cholPSig[t].T.dot(cholPSig[t])
        trajsig[t] = np.r_[
            np.c_[trajsig[t, ix, ix], trajsig[t, ix, ix].dot(K[t].T)],
            np.c_[K[t].dot(trajsig[t, ix, ix]), K[t].dot(trajsig[t, ix, ix]).dot(K[t].T) + PSig]
        ]
        cur_action = K[t].dot(mu[t, ix]) + k[t]
        mu[t] = np.r_[mu[t, ix], cur_action]
        # mu[t] = np.r_[mu[t,ix], np.zeros(dU)]

        # Reuse old dynamics
        if not time_varying_dynamics:
            if t == 0:
                F[0], f[0], dynsig[0] = dynamics.get_dynamics(cur_timestep, prevx, prevu, x0, cur_action)
            F[t] = F[0]
            f[t] = f[0]
            dynsig[t] = dynsig[0]

        if t < H:
            # Estimate new dynamics here based on mu
            if time_varying_dynamics and t < max_time_varying_horizon:
                F[t], f[t], dynsig[t] = dynamics.get_dynamics(cur_timestep + t, mu[t - 1, ix], mu[t - 1, iu],
                                                              mu[t, ix], cur_action)
        if t < H - 1:
            trajsig[t + 1, ix, ix] = F[t].dot(trajsig[t]).dot(F[t].T) + dynsig[t]
            mu[t + 1, ix] = F[t].dot(mu[t]) + f[t]

    # cc = np.zeros((N, H))
    cv = np.zeros((N, H, dT))
    Cm = np.zeros((N, H, dT, dT))

    for n in range(N):
        # Get costs.
        newmu = mu[:, iu]
        l, lx, lu, lxx, luu, lux = cost.eval(mu[:, ix], newmu, cur_timestep, jac=jacobian)

        cv[n] = np.c_[lx, lu]
        Cm[n] = np.concatenate((np.c_[lxx, np.transpose(lux, [0, 2, 1])], np.c_[lux, luu]), axis=1)

        # Adjust for expanding cost around a sample.
        yhat = mu  # np.c_[mu[:,ix], newmu_u]
        rdiff = -yhat  # T x (X+U)
        rdiff_expand = np.expand_dims(rdiff, axis=2)  # T x (X+U) x 1
        cv_update = np.sum(Cm[n] * rdiff_expand, axis=1)  # T x (X+U)
        cv[n] += cv_update

    cv = np.mean(cv, axis=0)
    Cm = np.mean(Cm, axis=0)
    return cv, Cm, F, f, mu, trajsig


def dlqr(A, B, Q, R):
    """Solve the discrete time lqr controller.

    x[k+1] = A x[k] + B u[k]

    cost = sum x[k].T*Q*x[k] + u[k].T*R*u[k]
    """

    # ref Bertsekas, p.151

    # first, try to solve the ricatti equation
    X = sp.linalg.solve_discrete_are(A, B, Q, R)

    # compute the LQR gain
    K = sp.linalg.inv(B.T @ X @ B + R) @ (B.T @ X @ A)

    eigVals, eigVecs = sp.linalg.eig(A - B @ K)

    return K, X, eigVals



import torch
from torch.autograd import Function
import numpy as np
import scipy.linalg
import logging

logger = logging.getLogger(__name__)


def batch_quadratic_product(X, A):
    """
    Batch multiplication of x^T A x.
    :param X: N x nx where each x is a row in X
    :param A: nx x nx
    :returns N x 1 product of each x^T A x
    """
    return torch.einsum('ij,kj,ik->i', X, A, X)


def batch_outer_product(u, v):
    """
    Batch outer product uv^T
    :param u: N x nx
    :param v: N x nx
    :return: N x nx x nx where each 'row' is the outer product of each row of u and v
    """
    return torch.einsum('ij,ik->ijk', u, v)


def batch_batch_product(X, A):
    """
    Batch multiplication of y = Ax where both A and x are batched
    :param X: N x nx
    :param A: N x ny x nx
    :return: N x ny product of each A with each x
    """
    return torch.einsum('ijk,ik->ij', A, X)


def kronecker_product(t1, t2):
    """
    Computes the Kronecker product between two tensors.
    See https://en.wikipedia.org/wiki/Kronecker_product
    """
    t1_height, t1_width = t1.size()
    t2_height, t2_width = t2.size()
    out_height = t1_height * t2_height
    out_width = t1_width * t2_width

    tiled_t2 = t2.repeat(t1_height, t1_width)
    expanded_t1 = (
        t1.unsqueeze(2)
            .unsqueeze(3)
            .repeat(1, t2_height, t2_width, 1)
            .view(out_height, out_width)
    )

    return expanded_t1 * tiled_t2


def cov(x, rowvar=False, bias=False, ddof=None, aweights=None):
    """Estimates covariance matrix like numpy.cov"""
    # ensure at least 2D
    if x.dim() == 1:
        x = x.view(-1, 1)

    # treat each column as a data point, each row as a variable
    if rowvar and x.shape[0] != 1:
        x = x.t()

    if ddof is None:
        if bias == 0:
            ddof = 1
        else:
            ddof = 0

    w = aweights
    if w is not None:
        if not torch.is_tensor(w):
            w = torch.tensor(w, dtype=torch.float)
        w_sum = torch.sum(w)
        avg = torch.sum(x * (w / w_sum)[:, None], 0)
    else:
        avg = torch.mean(x, 0)

    # Determine the normalization
    if w is None:
        fact = x.shape[0] - ddof
    elif ddof == 0:
        fact = w_sum
    elif aweights is None:
        fact = w_sum - ddof
    else:
        fact = w_sum - ddof * torch.sum(w * w) / w_sum

    xm = x.sub(avg.expand_as(x))

    if w is None:
        X_T = xm.t()
    else:
        X_T = torch.mm(torch.diag(w), xm).t()

    c = torch.mm(X_T, xm)
    c = c / fact

    return c.squeeze()


class MatrixSquareRoot(Function):
    """Square root of a positive definite matrix.
    From https://github.com/steveli/pytorch-sqrtm
    NOTE: matrix square root is not differentiable for matrices with
          zero eigenvalues.
    """

    @staticmethod
    def forward(ctx, input):
        m = input.detach().numpy().astype(np.float_)
        sqrtm = torch.from_numpy(scipy.linalg.sqrtm(m).real).type_as(input)
        ctx.save_for_backward(sqrtm)
        return sqrtm

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        if ctx.needs_input_grad[0]:
            sqrtm, = ctx.saved_tensors
            sqrtm = sqrtm.data.numpy().astype(np.float_)
            gm = grad_output.data.numpy().astype(np.float_)

            # Given a positive semi-definite matrix X,
            # since X = X^{1/2}X^{1/2}, we can compute the gradient of the
            # matrix square root dX^{1/2} by solving the Sylvester equation:
            # dX = (d(X^{1/2})X^{1/2} + X^{1/2}(dX^{1/2}).
            grad_sqrtm = scipy.linalg.solve_sylvester(sqrtm, sqrtm, gm)

            grad_input = torch.from_numpy(grad_sqrtm).type_as(grad_output.data)
        return grad_input


sqrtm = MatrixSquareRoot.apply


class GELS(Function):
    """ Efficient implementation of gels from
        Nanxin Chen
        bobchennan@gmail.com
    """

    @staticmethod
    def forward(ctx, A, b):
        # A: (..., M, N)
        # b: (..., M, K)
        # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/ops/linalg_ops.py#L267
        u = torch.cholesky(torch.matmul(A.transpose(-1, -2), A), upper=True)
        ret = torch.cholesky_solve(torch.matmul(A.transpose(-1, -2), b), u, upper=True)
        ctx.save_for_backward(u, ret, A, b)
        return ret

    @staticmethod
    def backward(ctx, grad_output):
        # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/ops/linalg_grad.py#L223
        chol, x, a, b = ctx.saved_tensors
        z = torch.cholesky_solve(grad_output, chol, upper=True)
        xzt = torch.matmul(x, z.transpose(-1, -2))
        zx_sym = xzt + xzt.transpose(-1, -2)
        grad_A = - torch.matmul(a, zx_sym) + torch.matmul(b, z.transpose(-1, -2))
        grad_b = torch.matmul(a, z)
        return grad_A, grad_b


def _apply_weights(X, Y, weights):
    if weights is not None:
        # assume errors are uncorrelated so weight is given as a vector rather than matrix
        w = torch.sqrt(weights).view(-1, 1)
        # multiplying a diagonal matrix is equal to multiplying each row by the corresponding weight
        X = w * X
        Y = w * Y
    return X, Y


def ls(X, Y, weights=None):
    X, Y = _apply_weights(X, Y, weights)

    # currently no gradient support for gels
    # params, _ = torch.gels(Y, X)
    params = GELS.apply(X, Y)
    # params = torch.solve(Y, X)
    # params = torch.mm(X.pinverse(), Y)

    return params


SYMMETRIC_CORRECTION_NORM_THRESHOLD = 1e-3


def ls_cov(X, Y, weights=None, make_symmetric=True, sigreg=1e-4):
    X, Y = _apply_weights(X, Y, weights)

    pinvXX = X.pinverse()
    params = (pinvXX @ Y).t()

    # estimate covariance according to: http://users.stat.umn.edu/~helwig/notes/mvlr-Notes.pdf (see up to slide 66)
    # hat/projection matrix - Yhat = H*Y
    H = X @ pinvXX

    N = X.shape[0]
    n = X.shape[1]
    # degrees of freedom
    dof = N - n
    assert dof >= 0
    # estimated error covariance (unbiased estimate of error cov matrix Sigma
    # regularize for ill-conditioned matrices
    a = torch.eye(N, dtype=H.dtype, device=X.device) - H
    error_covariance = Y.t() @ a @ Y / dof
    error_covariance += torch.eye(error_covariance.shape[0], dtype=H.dtype, device=X.device) * sigreg

    XXXX = X.t() @ X
    # regularize
    XXXX += torch.eye(XXXX.shape[0], dtype=H.dtype, device=X.device) * sigreg

    if make_symmetric:
        # correct to be symmetric (if needed)
        error_covariance_sym = (error_covariance + error_covariance.t()) / 2
        XXXX_sym = (XXXX + XXXX.t()) / 2

        e_correction_error = (error_covariance_sym - error_covariance).norm()
        x_correction_error = (XXXX_sym - XXXX).norm()
        if e_correction_error > SYMMETRIC_CORRECTION_NORM_THRESHOLD or x_correction_error > SYMMETRIC_CORRECTION_NORM_THRESHOLD:
            logger.debug('unsymmetric covariances with error %f and %f', e_correction_error, x_correction_error)
        XXXX = XXXX_sym
        error_covariance = error_covariance_sym

    # TODO might be able to use cholesky decomp here since XXXX > 0
    covariance = kronecker_product(error_covariance, XXXX.inverse())

    return params, covariance


if __name__ == "__main__":
    d = torch.tensor([[1, 2], [3, 4]], dtype=torch.float)
    e = torch.eye(3, dtype=torch.float)
    f = kronecker_product(d.t(), e)
    print(f)



import torch
import functools
import numpy as np


def is_tensor_like(x):
    return torch.is_tensor(x) or type(x) is np.ndarray


def ensure_2d_input(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args = [v.reshape(1, -1) if (is_tensor_like(v) and len(v.shape) == 1) else v for v in args]
        return func(*args, **kwargs)

    return wrapper


def handle_batch_input(func):
    """For func that expect 2D input, handle input that have more than 2 dimensions by flattening them temporarily"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # assume inputs that are tensor-like have compatible shapes and is represented by the first argument
        batch_dims = []
        for arg in args:
            if is_tensor_like(arg) and len(arg.shape) > 2:
                batch_dims = arg.shape[:-1]  # last dimension is type dependent; all previous ones are batches
                break
        # no batches; just return normally
        if not batch_dims:
            return func(*args, **kwargs)

        # reduce all batch dimensions down to the first one
        args = [v.view(-1, v.shape[-1]) if (is_tensor_like(v) and len(v.shape) > 2) else v for v in args]
        ret = func(*args, **kwargs)
        # restore original batch dimensions; keep variable dimension (nx)
        if type(ret) is tuple:
            ret = [v if (not is_tensor_like(v) or len(v.shape) == 0) else (
                v.view(*batch_dims, v.shape[-1]) if len(v.shape) == 2 else v.view(*batch_dims)) for v in ret]
        else:
            if is_tensor_like(ret):
                if len(ret.shape) == 2:
                    ret = ret.view(*batch_dims, ret.shape[-1])
                else:
                    ret = ret.view(*batch_dims)
        return ret

    return wrapper


def ensure_tensor(device, dtype, *args):
    tensors = tuple(
        x.to(device=device, dtype=dtype) if torch.is_tensor(x) else
        torch.tensor(x, device=device, dtype=dtype)
        for x in args)
    return tensors if len(tensors) > 1 else tensors[0]


def ensure_diagonal(Q, dim):
    """
    Ensure that Q is a (dim x dim) diagonal matrix either with values of Q, or with eye * Q if Q is scalar
    :param Q: (dim x dim) tensor or numpy array, or a (dim) vector of the diagonal, or a scalar to scale an identity
    matrix by
    :param dim: size of matrix
    :return: (dim x dim) diagonal matrix
    """
    if not torch.is_tensor(Q):
        if type(Q) in (int, float):
            Q = torch.eye(dim) * Q
        else:
            Q = torch.tensor(Q)
    if len(Q.shape) == 1:
        if Q.shape[0] is not dim:
            raise RuntimeError("Expect {} sized diagonal vector but given {}".format(dim, Q.shape[0]))
        Q = torch.diag(Q)
    return Q



import torch
from arm_pytorch_utilities import tensor_utils


def test_ensure_diagonal():
    nx = 5
    q = torch.rand(nx)
    Q = torch.diag(q)
    Q1 = tensor_utils.ensure_diagonal(Q, nx)
    assert torch.allclose(Q, Q1)
    Q2 = tensor_utils.ensure_diagonal(q, nx)
    assert torch.allclose(Q, Q2)
    Q3 = tensor_utils.ensure_diagonal(q.numpy(), nx)
    assert torch.allclose(Q, Q3)
    Q4 = tensor_utils.ensure_diagonal(Q.numpy(), nx)
    assert torch.allclose(Q, Q4)

    q = torch.rand(1).item()
    Q = torch.eye(nx) * q
    Q1 = tensor_utils.ensure_diagonal(q, nx)
    assert torch.allclose(Q, Q1)


def test_handle_batch_input():
    @tensor_utils.handle_batch_input
    def add_and_average(a, b):
        assert len(a.shape) == 2
        return a + b, (a + b).mean()

    B = 4
    N = 10
    nx = 3
    A = torch.rand((B, N, nx))
    Ahat, mean = add_and_average(A, 0)

    assert A.shape == Ahat.shape
    assert torch.allclose(A, Ahat)


if __name__ == "__main__":
    test_ensure_diagonal()
    test_handle_batch_input()
